from fractions import Fraction
import logging
import os

from av.bitstream.context cimport BitStreamFilterContext
from av.codec.codec cimport Codec
from av.codec.context cimport CodecContext, wrap_codec_context
from av.container.streams cimport StreamContainer
from av.dictionary cimport _Dictionary
from av.error cimport err_check
from av.packet cimport Packet
from av.stream cimport Stream, wrap_stream
from av.utils cimport dict_to_avdict, to_avrational

from av.dictionary import Dictionary


log = logging.getLogger(__name__)


cdef close_output(OutputContainer self):
    if self._started and not self._done:
        # We must only ever call av_write_trailer *once*, otherwise we get a
        # segmentation fault. Therefore no matter whether it succeeds or not
        # we must absolutely set self._done.
        try:
            self.err_check(lib.av_write_trailer(self.ptr))
        finally:
            if self.file is None and not (self.ptr.oformat.flags & lib.AVFMT_NOFILE):
                lib.avio_closep(&self.ptr.pb)
            self._done = True


cdef class OutputContainer(Container):

    def __cinit__(self, *args, **kwargs):
        self.streams = StreamContainer()
        self.metadata = {}
        with nogil:
            self.packet_ptr = lib.av_packet_alloc()

    def __dealloc__(self):
        close_output(self)
        with nogil:
            lib.av_packet_free(&self.packet_ptr)

    def add_stream(self, codec_name=None, object rate=None, object template=None, options=None, **kwargs):
        """add_stream(codec_name, rate=None)

        Create a new stream, and return it.

        :param str codec_name: The name of a codec.
        :param rate: The frame rate for video, and sample rate for audio.
            Examples for video include ``24``, ``23.976``, and
            ``Fraction(30000,1001)``. Examples for audio include ``48000``
            and ``44100``.
        :returns: The new :class:`~av.stream.Stream`.

        """

        if not (template is None or isinstance(template, (Stream, BitStreamFilterContext))):
            raise ValueError("template must be Stream or BitStreamFilterContext")
        if (codec_name is None and template is None) or (codec_name is not None and template is not None):
            raise ValueError('needs one of codec_name or template')

        cdef const lib.AVCodec *codec
        cdef Codec codec_obj

        if codec_name is not None:
            codec_obj = codec_name if isinstance(codec_name, Codec) else Codec(codec_name, 'w')
        elif isinstance(template, BitStreamFilterContext):
            codec_obj = Codec(lib.avcodec_get_name((<BitStreamFilterContext>template).ptr.par_out.codec_id))
        else:
            if not template.codec_context:
                raise ValueError("template has no codec context")
            codec_obj = (<Stream>template).codec_context.codec
        codec = codec_obj.ptr

        # Assert that this format supports the requested codec.
        if not lib.avformat_query_codec(
            self.ptr.oformat,
            codec.id,
            lib.FF_COMPLIANCE_NORMAL,
        ):
            raise ValueError("%r format does not support %r codec" % (self.format.name, codec_name))

        # Create new stream in the AVFormatContext, set AVCodecContext values.
        lib.avformat_new_stream(self.ptr, codec)
        cdef lib.AVStream *stream = self.ptr.streams[self.ptr.nb_streams - 1]
        cdef lib.AVCodecContext *codec_context = lib.avcodec_alloc_context3(codec)

        # Copy from the template.
        if template is not None:
            if isinstance(template, BitStreamFilterContext):
                err_check(lib.avcodec_parameters_to_context(codec_context, (<BitStreamFilterContext>template).ptr.par_out))
                stream.time_base = codec_context.pkt_timebase = (<BitStreamFilterContext>template).ptr.time_base_out
            else:
                err_check(lib.avcodec_parameters_to_context(codec_context, (<Stream>template).ptr.codecpar))
            # Reset the codec tag assuming we are remuxing.
            codec_context.codec_tag = 0

        # Now lets set some more sane video defaults
        elif codec.type == lib.AVMEDIA_TYPE_VIDEO:
            codec_context.pix_fmt = lib.AV_PIX_FMT_YUV420P
            codec_context.width = 640
            codec_context.height = 480
            codec_context.bit_rate = 1024000
            codec_context.bit_rate_tolerance = 128000
            codec_context.ticks_per_frame = 1
            to_avrational(rate or 24, &codec_context.framerate)

            stream.avg_frame_rate = codec_context.framerate
            stream.time_base = codec_context.time_base

        # Some sane audio defaults
        elif codec.type == lib.AVMEDIA_TYPE_AUDIO:
            codec_context.sample_fmt = codec.sample_fmts[0]
            codec_context.bit_rate = 128000
            codec_context.bit_rate_tolerance = 32000
            codec_context.sample_rate = rate or 48000
            codec_context.channels = 2
            codec_context.channel_layout = lib.AV_CH_LAYOUT_STEREO

        # Some formats want stream headers to be separate
        if self.ptr.oformat.flags & lib.AVFMT_GLOBALHEADER:
            codec_context.flags |= lib.AV_CODEC_FLAG_GLOBAL_HEADER

        # Initialise stream codec parameters to populate the codec type.
        #
        # Subsequent changes to the codec context will be applied just before
        # encoding starts in `start_encoding()`.
        err_check(lib.avcodec_parameters_from_context(stream.codecpar, codec_context))

        # Construct the user-land stream
        cdef CodecContext py_codec_context = wrap_codec_context(codec_context, codec)
        cdef Stream py_stream = wrap_stream(self, stream, py_codec_context)
        self.streams.add_stream(py_stream)

        if options:
            py_stream.options.update(options)

        for k, v in kwargs.items():
            setattr(py_stream, k, v)

        return py_stream

    cpdef start_encoding(self):
        """Write the file header! Called automatically."""

        if self._started:
            return

        # TODO: This does NOT handle options coming from 3 sources.
        # This is only a rough approximation of what would be cool to do.
        used_options = set()

        # Finalize and open all streams.
        cdef Stream stream
        for stream in self.streams:

            ctx = stream.codec_context
            if not ctx.is_open:

                for k, v in self.options.items():
                    ctx.options.setdefault(k, v)
                ctx.open()

                # Track option consumption.
                for k in self.options:
                    if k not in ctx.options:
                        used_options.add(k)

            stream._finalize_for_output()

        # Open the output file, if needed.
        cdef bytes name_obj = os.fsencode(self.name if self.file is None else "")
        cdef char *name = name_obj
        if self.ptr.pb == NULL and not self.ptr.oformat.flags & lib.AVFMT_NOFILE:
            err_check(lib.avio_open(&self.ptr.pb, name, lib.AVIO_FLAG_WRITE))

        # Copy the metadata dict.
        dict_to_avdict(
            &self.ptr.metadata, self.metadata,
            encoding=self.metadata_encoding,
            errors=self.metadata_errors
        )

        cdef _Dictionary all_options = Dictionary(self.options, self.container_options)
        cdef _Dictionary options = all_options.copy()
        self.err_check(lib.avformat_write_header(
            self.ptr,
            &options.ptr
        ))

        # Track option usage...
        for k in all_options:
            if k not in options:
                used_options.add(k)
        # ... and warn if any weren't used.
        unused_options = {k: v for k, v in self.options.items() if k not in used_options}
        if unused_options:
            log.warning('Some options were not used: %s' % unused_options)

        self._started = True

    def close(self):
        for stream in self.streams:
            if stream.codec_context:
                stream.codec_context.close(strict=False)

        close_output(self)

    def mux(self, packets):
        # We accept either a Packet, or a sequence of packets. This should
        # smooth out the transition to the new encode API which returns a
        # sequence of packets.
        if isinstance(packets, Packet):
            self.mux_one(packets)
        else:
            for packet in packets:
                self.mux_one(packet)

    def mux_one(self, Packet packet not None):
        self.start_encoding()

        # Assert the packet is in stream time.
        if packet.ptr.stream_index < 0 or <unsigned int>packet.ptr.stream_index >= self.ptr.nb_streams:
            raise ValueError('Bad Packet stream_index.')
        cdef lib.AVStream *stream = self.ptr.streams[packet.ptr.stream_index]
        packet._rebase_time(stream.time_base)

        # Make another reference to the packet, as av_interleaved_write_frame
        # takes ownership of the reference.
        self.err_check(lib.av_packet_ref(self.packet_ptr, packet.ptr))

        cdef int ret
        with nogil:
            ret = lib.av_interleaved_write_frame(self.ptr, self.packet_ptr)
        self.err_check(ret)



"""

PyAV provides enumeration and flag classes that are similar to the stdlib ``enum``
module that shipped with Python 3.4.

PyAV's enums are a little more forgiving to preserve backwards compatibility
with earlier PyAV patterns. e.g., they can be freely compared to strings or
integers for names and values respectively.

"""

from collections import OrderedDict
import copyreg
import sys


cdef sentinel = object()


class EnumType(type):

    def __new__(mcl, name, bases, attrs, *args):
        # Just adapting the method signature.
        return super().__new__(mcl, name, bases, attrs)

    def __init__(self, name, bases, attrs, items):

        self._by_name = {}
        self._by_value = {}
        self._all = []

        for spec in items:
            self._create(*spec)

    def _create(self, name, value, doc=None, by_value_only=False):

        # We only have one instance per value.
        try:
            item = self._by_value[value]
        except KeyError:
            item = self(sentinel, name, value, doc)
            self._by_value[value] = item

        if not by_value_only:
            setattr(self, name, item)
            self._all.append(item)
            self._by_name[name] = item

        return item

    def __len__(self):
        return len(self._all)

    def __iter__(self):
        return iter(self._all)

    def __getitem__(self, key):

        if isinstance(key, str):
            return self._by_name[key]

        if isinstance(key, int):
            try:
                return self._by_value[key]
            except KeyError:
                pass

            if issubclass(self, EnumFlag):
                return self._get_multi_flags(key)

            raise KeyError(key)

        if isinstance(key, self):
            return key

        raise TypeError("{0} indices must be str, int, or {0}".format(
            self.__name__,
        ))

    def _get(self, long value, bint create=False):

        try:
            return self._by_value[value]
        except KeyError:
            pass

        if not create:
            return

        return self._create('{}_{}'.format(self.__name__.upper(), value), value, by_value_only=True)

    def _get_multi_flags(self, long value):

        try:
            return self._by_value[value]
        except KeyError:
            pass

        flags = []
        cdef long to_find = value
        for item in self:
            if item.value & to_find:
                flags.append(item)
                to_find = to_find ^ item.value
                if not to_find:
                    break

        if to_find:
            raise KeyError(value)

        name = '|'.join(f.name for f in flags)
        cdef EnumFlag combo = self._create(name, value, by_value_only=True)
        combo.flags = tuple(flags)

        return combo

    def get(self, key, default=None, create=False):
        try:
            return self[key]
        except KeyError:
            if create:
                return self._get(key, create=True)
            return default

    def property(self, *args, **kwargs):
        return EnumProperty(self, *args, **kwargs)


def _unpickle(mod_name, cls_name, item_name):
    mod = __import__(mod_name, fromlist=['.'])
    cls = getattr(mod, cls_name)
    return cls[item_name]


copyreg.constructor(_unpickle)


cdef class EnumItem(object):

    """
    Enumerations are when an attribute may only take on a single value at once, and
    they are represented as integers in the FFmpeg API. We associate names with each
    value that are easier to operate with.

    Consider :data:`av.codec.context.SkipType`, which is the type of the :attr:`CodecContext.skip_frame` attribute::

        >>> fh = av.open(video_path)
        >>> cc = fh.streams.video[0].codec_context

        >>> # The skip_frame attribute has a name and value:
        >>> cc.skip_frame.name
        'DEFAULT'
        >>> cc.skip_frame.value
        0

        >>> # You can compare it to strings and ints:
        >>> cc.skip_frame == 'DEFAULT'
        True
        >>> cc.skip_frame == 0
        True

        >>> # You can assign strings and ints:
        >>> cc.skip_frame = 'NONKEY'
        >>> cc.skip_frame == 'NONKEY'
        True
        >>> cc.skip_frame == 32
        True

    """

    cdef readonly str name
    cdef readonly int value
    cdef Py_hash_t _hash

    def __cinit__(self, sentinel_, str name, int value, doc=None):

        if sentinel_ is not sentinel:
            raise RuntimeError("Cannot instantiate {}.".format(self.__class__.__name__))

        self.name = name
        self.value = value
        self.__doc__ = doc  # This is not cdef because it doesn't work if it is.

        # We need to establish a hash that doesn't collide with anything that
        # would return true from `__eq__`. This is because these enums (vs
        # the stdlib ones) are weakly typed (they will compare against string
        # names and int values), and if we have the same hash AND are equal,
        # then they will be equivalent as keys in a dictionary, which is wierd.
        cdef Py_hash_t hash_ = value + 1
        if hash_ == hash(name):
            hash_ += 1
        self._hash = hash_

    def __repr__(self):
        return '<{}.{}:{}(0x{:x})>'.format(
            self.__class__.__module__,
            self.__class__.__name__,
            self.name,
            self.value,
        )

    def __str__(self):
        return self.name

    def __int__(self):
        return self.value

    def __hash__(self):
        return self._hash

    def __reduce__(self):
        return (_unpickle, (self.__class__.__module__, self.__class__.__name__, self.name))

    def __eq__(self, other):

        if isinstance(other, str):

            if self.name == other:  # The quick method.
                return True

            try:
                other_inst = self.__class__._by_name[other]
            except KeyError:
                raise ValueError("{} does not have item named {!r}".format(
                    self.__class__.__name__,
                    other,
                ))
            else:
                return self is other_inst

        if isinstance(other, int):
            if self.value == other:
                return True
            if other in self.__class__._by_value:
                return False
            raise ValueError("{} does not have item valued {}".format(
                self.__class__.__name__,
                other,
            ))

        if isinstance(other, self.__class__):
            return self is other

        raise TypeError("'==' not supported between {} and {}".format(
            self.__class__.__name__,
            type(other).__name__,
        ))

    def __ne__(self, other):
        return not (self == other)


cdef class EnumFlag(EnumItem):

    """
    Flags are sets of boolean attributes, which the FFmpeg API represents as individual
    bits in a larger integer which you manipulate with the bitwise operators.
    We associate names with each flag that are easier to operate with.

    Consider :data:`CodecContextFlags`, whis is the type of the :attr:`CodecContext.flags`
    attribute, and the set of boolean properties::

        >>> fh = av.open(video_path)
        >>> cc = fh.streams.video[0].codec_context

        >>> cc.flags
        <av.codec.context.Flags:NONE(0x0)>

        >>> # You can set flags via bitwise operations with the objects, names, or values:
        >>> cc.flags |= cc.flags.OUTPUT_CORRUPT
        >>> cc.flags |= 'GLOBAL_HEADER'
        >>> cc.flags
        <av.codec.context.Flags:OUTPUT_CORRUPT|GLOBAL_HEADER(0x400008)>

        >>> # You can test flags via bitwise operations with objects, names, or values:
        >>> bool(cc.flags & cc.flags.OUTPUT_CORRUPT)
        True
        >>> bool(cc.flags & 'QSCALE')
        False

        >>> # There are boolean properties for each flag:
        >>> cc.output_corrupt
        True
        >>> cc.qscale
        False

        >>> # You can set them:
        >>> cc.qscale = True
        >>> cc.flags
        <av.codec.context.Flags:QSCALE|OUTPUT_CORRUPT|GLOBAL_HEADER(0x40000a)>

    """

    cdef readonly tuple flags

    def __cinit__(self, sentinel, name, value, doc=None):
        self.flags = (self, )

    def __and__(self, other):
        if not isinstance(other, int):
            other = self.__class__[other].value
        value = self.value & other
        return self.__class__._get_multi_flags(value)

    def __or__(self, other):
        if not isinstance(other, int):
            other = self.__class__[other].value
        value = self.value | other
        return self.__class__._get_multi_flags(value)

    def __xor__(self, other):
        if not isinstance(other, int):
            other = self.__class__[other].value
        value = self.value ^ other
        return self.__class__._get_multi_flags(value)

    def __invert__(self):
        # This can't result in a flag, but is helpful.
        return ~self.value

    def __nonzero__(self):
        return bool(self.value)


cdef class EnumProperty(object):

    cdef object enum
    cdef object fget
    cdef object fset
    cdef public __doc__

    def __init__(self, enum, fget, fset=None, doc=None):
        self.enum = enum
        self.fget = fget
        self.fset = fset
        self.__doc__ = doc or fget.__doc__

    def setter(self, fset):
        self.fset = fset
        return self

    def __get__(self, inst, owner):
        if inst is not None:
            value = self.fget(inst)
            return self.enum.get(value, create=True)
        else:
            return self

    def __set__(self, inst, value):
        item = self.enum.get(value)
        self.fset(inst, item.value)

    def flag_property(self, name, doc=None):

        item = self.enum[name]
        cdef int item_value = item.value

        class Property(property):
            pass

        @Property
        def _property(inst):
            return bool(self.fget(inst) & item_value)

        if self.fset:
            @_property.setter
            def _property(inst, value):
                if value:
                    flags = self.fget(inst) | item_value
                else:
                    flags = self.fget(inst) & ~item_value
                self.fset(inst, flags)

        _property.__doc__ = doc or item.__doc__
        _property._enum_item = item

        return _property


cpdef define_enum(name, module, items, bint is_flags=False):

    if is_flags:
        base_cls = EnumFlag
    else:
        base_cls = EnumItem

    # Some items may be None if they correspond to an unsupported FFmpeg feature
    cls = EnumType(name, (base_cls, ), {'__module__': module}, [i for i in items if i is not None])

    return cls



from av import ContainerFormat, formats_available

from .common import TestCase


class TestContainerFormats(TestCase):
    def test_matroska(self):
        fmt = ContainerFormat("matroska")
        self.assertTrue(fmt.is_input)
        self.assertTrue(fmt.is_output)
        self.assertEqual(fmt.name, "matroska")
        self.assertEqual(fmt.long_name, "Matroska")
        self.assertIn("mkv", fmt.extensions)
        self.assertFalse(fmt.no_file)

    def test_mov(self):
        fmt = ContainerFormat("mov")
        self.assertTrue(fmt.is_input)
        self.assertTrue(fmt.is_output)
        self.assertEqual(fmt.name, "mov")
        self.assertEqual(fmt.long_name, "QuickTime / MOV")
        self.assertIn("mov", fmt.extensions)
        self.assertFalse(fmt.no_file)

    def test_stream_segment(self):
        # This format goes by two names, check both.
        fmt = ContainerFormat("stream_segment")
        self.assertFalse(fmt.is_input)
        self.assertTrue(fmt.is_output)
        self.assertEqual(fmt.name, "stream_segment")
        self.assertEqual(fmt.long_name, "streaming segment muxer")
        self.assertEqual(fmt.extensions, set())
        self.assertTrue(fmt.no_file)

        fmt = ContainerFormat("ssegment")
        self.assertFalse(fmt.is_input)
        self.assertTrue(fmt.is_output)
        self.assertEqual(fmt.name, "ssegment")
        self.assertEqual(fmt.long_name, "streaming segment muxer")
        self.assertEqual(fmt.extensions, set())
        self.assertTrue(fmt.no_file)

    def test_formats_available(self):
        self.assertTrue(formats_available)




cdef extern from "libavfilter/avfilter.h" nogil:
    """
    #if (LIBAVFILTER_VERSION_INT >= 525156)
        // avfilter_filter_pad_count is available since version 8.3.100 of libavfilter (FFmpeg 5.0)
        #define _avfilter_get_num_pads(filter, is_output, pads) (avfilter_filter_pad_count(filter, is_output))
    #else
        // avfilter_filter_pad_count has been deprecated as of version 8.3.100 of libavfilter (FFmpeg 5.0)
        #define _avfilter_get_num_pads(filter, is_output, pads) (avfilter_pad_count(pads))
    #endif
    """
    cdef int   avfilter_version()
    cdef char* avfilter_configuration()
    cdef char* avfilter_license()

    cdef struct AVFilterPad:
        # This struct is opaque.
        pass

    const char* avfilter_pad_get_name(const AVFilterPad *pads, int index)
    AVMediaType avfilter_pad_get_type(const AVFilterPad *pads, int index)

    int pyav_get_num_pads "_avfilter_get_num_pads" (const AVFilter *filter, int is_output, const AVFilterPad *pads)

    cdef struct AVFilter:

        AVClass *priv_class

        const char *name
        const char *description

        const int flags

        const AVFilterPad *inputs
        const AVFilterPad *outputs
        int (*process_command)(AVFilterContext *, const char *cmd, const char *arg, char *res, int res_len, int flags)

    cdef enum:
        AVFILTER_FLAG_DYNAMIC_INPUTS
        AVFILTER_FLAG_DYNAMIC_OUTPUTS
        AVFILTER_FLAG_SLICE_THREADS
        AVFILTER_FLAG_SUPPORT_TIMELINE_GENERIC
        AVFILTER_FLAG_SUPPORT_TIMELINE_INTERNAL

    cdef AVFilter* avfilter_get_by_name(const char *name)
    cdef const AVFilter* av_filter_iterate(void **opaque)

    cdef struct AVFilterLink # Defined later.

    cdef struct AVFilterContext:

        AVClass *av_class
        AVFilter *filter

        char *name

        unsigned int nb_inputs
        AVFilterPad *input_pads
        AVFilterLink **inputs

        unsigned int nb_outputs
        AVFilterPad *output_pads
        AVFilterLink **outputs

    cdef int avfilter_init_str(AVFilterContext *ctx, const char *args)
    cdef int avfilter_init_dict(AVFilterContext *ctx, AVDictionary **options)
    cdef void avfilter_free(AVFilterContext*)
    cdef AVClass* avfilter_get_class()

    cdef struct AVFilterLink:

        AVFilterContext *src
        AVFilterPad *srcpad
        AVFilterContext *dst
        AVFilterPad *dstpad

        AVMediaType Type
        int w
        int h
        AVRational sample_aspect_ratio
        uint64_t channel_layout
        int sample_rate
        int format
        AVRational time_base

    # custom
    cdef set pyav_get_available_filters()


cdef extern from "libavfilter/buffersink.h" nogil:
    cdef void av_buffersink_set_frame_size(AVFilterContext *ctx, unsigned frame_size)



cimport libav as lib

from av.audio.format cimport AudioFormat, get_audio_format
from av.audio.frame cimport AudioFrame, alloc_audio_frame
from av.audio.layout cimport AudioLayout, get_audio_layout
from av.error cimport err_check
from av.frame cimport Frame
from av.packet cimport Packet


cdef class AudioCodecContext(CodecContext):

    cdef _init(self, lib.AVCodecContext *ptr, const lib.AVCodec *codec):
        CodecContext._init(self, ptr, codec)

        # Sometimes there isn't a layout set, but there are a number of
        # channels. Assume it is the default layout.
        # TODO: Put this behind `not bare_metal`.
        # TODO: Do this more efficiently.
        if self.ptr.channels and not self.ptr.channel_layout:
            self.ptr.channel_layout = get_audio_layout(self.ptr.channels, 0).layout

    cdef _set_default_time_base(self):
        self.ptr.time_base.num = 1
        self.ptr.time_base.den = self.ptr.sample_rate

    cdef _prepare_frames_for_encode(self, Frame input_frame):

        cdef AudioFrame frame = input_frame

        cdef bint allow_var_frame_size = self.ptr.codec.capabilities & lib.AV_CODEC_CAP_VARIABLE_FRAME_SIZE

        # Note that the resampler will simply return an input frame if there is
        # no resampling to be done. The control flow was just a little easier this way.
        if not self.resampler:
            self.resampler = AudioResampler(
                format=self.format,
                layout=self.layout,
                rate=self.ptr.sample_rate,
                frame_size=None if allow_var_frame_size else self.ptr.frame_size
            )
        frames = self.resampler.resample(frame)

        # flush if input frame is None
        if input_frame is None:
            frames.append(None)

        return frames

    cdef Frame _alloc_next_frame(self):
        return alloc_audio_frame()

    cdef _setup_decoded_frame(self, Frame frame, Packet packet):
        CodecContext._setup_decoded_frame(self, frame, packet)
        cdef AudioFrame aframe = frame
        aframe._init_user_attributes()

    property frame_size:
        """
        Number of samples per channel in an audio frame.

        :type: int
        """
        def __get__(self): return self.ptr.frame_size

    property sample_rate:
        """
        Sample rate of the audio data, in samples per second.

        :type: int
        """
        def __get__(self):
            return self.ptr.sample_rate

        def __set__(self, int value):
            self.ptr.sample_rate = value

    property rate:
        """Another name for :attr:`sample_rate`."""
        def __get__(self):
            return self.sample_rate

        def __set__(self, value):
            self.sample_rate = value

    # TODO: Integrate into AudioLayout.
    property channels:
        def __get__(self):
            return self.ptr.channels

        def __set__(self, value):
            self.ptr.channels = value
            self.ptr.channel_layout = lib.av_get_default_channel_layout(value)
    property channel_layout:
        def __get__(self):
            return self.ptr.channel_layout

    property layout:
        """
        The audio channel layout.

        :type: AudioLayout
        """
        def __get__(self):
            return get_audio_layout(self.ptr.channels, self.ptr.channel_layout)

        def __set__(self, value):
            cdef AudioLayout layout = AudioLayout(value)
            self.ptr.channel_layout = layout.layout
            self.ptr.channels = layout.nb_channels

    property format:
        """
        The audio sample format.

        :type: AudioFormat
        """
        def __get__(self):
            return get_audio_format(self.ptr.sample_fmt)

        def __set__(self, value):
            cdef AudioFormat format = AudioFormat(value)
            self.ptr.sample_fmt = format.sample_fmt



import sys


cdef str container_format_postfix = 'le' if sys.byteorder == 'little' else 'be'


cdef object _cinit_bypass_sentinel

cdef AudioFormat get_audio_format(lib.AVSampleFormat c_format):
    """Get an AudioFormat without going through a string."""

    if c_format < 0:
        return None

    cdef AudioFormat format = AudioFormat.__new__(AudioFormat, _cinit_bypass_sentinel)
    format._init(c_format)
    return format


cdef class AudioFormat(object):

    """Descriptor of audio formats."""

    def __cinit__(self, name):

        if name is _cinit_bypass_sentinel:
            return

        cdef lib.AVSampleFormat sample_fmt
        if isinstance(name, AudioFormat):
            sample_fmt = (<AudioFormat>name).sample_fmt
        else:
            sample_fmt = lib.av_get_sample_fmt(name)

        if sample_fmt < 0:
            raise ValueError('Not a sample format: %r' % name)

        self._init(sample_fmt)

    cdef _init(self, lib.AVSampleFormat sample_fmt):
        self.sample_fmt = sample_fmt

    def __repr__(self):
        return '<av.AudioFormat %s>' % (self.name)

    property name:
        """Canonical name of the sample format.

        >>> SampleFormat('s16p').name
        's16p'

        """
        def __get__(self):
            return <str>lib.av_get_sample_fmt_name(self.sample_fmt)

    property bytes:
        """Number of bytes per sample.

        >>> SampleFormat('s16p').bytes
        2

        """
        def __get__(self):
            return lib.av_get_bytes_per_sample(self.sample_fmt)

    property bits:
        """Number of bits per sample.

        >>> SampleFormat('s16p').bits
        16

        """
        def __get__(self):
            return lib.av_get_bytes_per_sample(self.sample_fmt) << 3

    property is_planar:
        """Is this a planar format?

        Strictly opposite of :attr:`is_packed`.

        """
        def __get__(self):
            return bool(lib.av_sample_fmt_is_planar(self.sample_fmt))

    property is_packed:
        """Is this a planar format?

        Strictly opposite of :attr:`is_planar`.

        """
        def __get__(self):
            return not lib.av_sample_fmt_is_planar(self.sample_fmt)

    property planar:
        """The planar variant of this format.

        Is itself when planar:

        >>> fmt = Format('s16p')
        >>> fmt.planar is fmt
        True

        """
        def __get__(self):
            if self.is_planar:
                return self
            return get_audio_format(lib.av_get_planar_sample_fmt(self.sample_fmt))

    property packed:
        """The packed variant of this format.

        Is itself when packed:

        >>> fmt = Format('s16')
        >>> fmt.packed is fmt
        True

        """
        def __get__(self):
            if self.is_packed:
                return self
            return get_audio_format(lib.av_get_packed_sample_fmt(self.sample_fmt))

    property container_name:
        """The name of a :class:`ContainerFormat` which directly accepts this data.

        :raises ValueError: when planar, since there are no such containers.

        """
        def __get__(self):

            if self.is_planar:
                raise ValueError('no planar container formats')

            if self.sample_fmt == lib.AV_SAMPLE_FMT_U8:
                return 'u8'
            elif self.sample_fmt == lib.AV_SAMPLE_FMT_S16:
                return 's16' + container_format_postfix
            elif self.sample_fmt == lib.AV_SAMPLE_FMT_S32:
                return 's32' + container_format_postfix
            elif self.sample_fmt == lib.AV_SAMPLE_FMT_FLT:
                return 'f32' + container_format_postfix
            elif self.sample_fmt == lib.AV_SAMPLE_FMT_DBL:
                return 'f64' + container_format_postfix

            raise ValueError('unknown layout')



from collections.abc import Sequence


cdef object _cinit_bypass_sentinel = object()


# Cython doesn't let us inherit from the abstract Sequence, so we will subclass
# it later.
cdef class _MotionVectors(SideData):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._vectors = {}
        self._len = self.ptr.size // sizeof(lib.AVMotionVector)

    def __repr__(self):
        return f'<av.sidedata.MotionVectors {self.ptr.size} bytes of {len(self)} vectors at 0x{<unsigned int>self.ptr.data:0x}'

    def __getitem__(self, int index):

        try:
            return self._vectors[index]
        except KeyError:
            pass

        if index >= self._len:
            raise IndexError(index)

        vector = self._vectors[index] = MotionVector(_cinit_bypass_sentinel, self, index)
        return vector

    def __len__(self):
        return self._len

    def to_ndarray(self):
        import numpy as np
        return np.frombuffer(self, dtype=np.dtype([
            ('source', 'int32'),
            ('w', 'uint8'),
            ('h', 'uint8'),
            ('src_x', 'int16'),
            ('src_y', 'int16'),
            ('dst_x', 'int16'),
            ('dst_y', 'int16'),
            ('flags', 'uint64'),
            ('motion_x', 'int32'),
            ('motion_y', 'int32'),
            ('motion_scale', 'uint16'),
        ], align=True))


class MotionVectors(_MotionVectors, Sequence):
    pass


cdef class MotionVector(object):

    def __init__(self, sentinel, _MotionVectors parent, int index):
        if sentinel is not _cinit_bypass_sentinel:
            raise RuntimeError('cannot manually instatiate MotionVector')
        self.parent = parent
        cdef lib.AVMotionVector *base = <lib.AVMotionVector*>parent.ptr.data
        self.ptr = base + index

    def __repr__(self):
        return f'<av.sidedata.MotionVector {self.w}x{self.h} from {self.src_x},{self.src_y} to {self.dst_x},{self.dst_y}>'

    @property
    def source(self):
        return self.ptr.source

    @property
    def w(self):
        return self.ptr.w

    @property
    def h(self):
        return self.ptr.h

    @property
    def src_x(self):
        return self.ptr.src_x

    @property
    def src_y(self):
        return self.ptr.src_y

    @property
    def dst_x(self):
        return self.ptr.dst_x

    @property
    def dst_y(self):
        return self.ptr.dst_y

    @property
    def motion_x(self):
        return self.ptr.motion_x

    @property
    def motion_y(self):
        return self.ptr.motion_y

    @property
    def motion_scale(self):
        return self.ptr.motion_scale



cdef extern from "libavutil/dict.h" nogil:

    # See: http://ffmpeg.org/doxygen/trunk/structAVDictionary.html
    ctypedef struct AVDictionary:
        pass

    cdef void av_dict_free(AVDictionary **)

    # See: http://ffmpeg.org/doxygen/trunk/structAVDictionaryEntry.html
    ctypedef struct AVDictionaryEntry:
        char *key
        char *value

    cdef int AV_DICT_IGNORE_SUFFIX

    cdef AVDictionaryEntry* av_dict_get(
        AVDictionary *dict,
        char *key,
        AVDictionaryEntry *prev,
        int flags,
    )

    cdef int av_dict_set(
        AVDictionary **pm,
        const char *key,
        const char *value,
        int flags
    )

    cdef int av_dict_count(
        AVDictionary *m
    )

    cdef int av_dict_copy(
        AVDictionary **dst,
        AVDictionary *src,
        int flags
    )




cdef class Plane(Buffer):
    """
    Base class for audio and video planes.

    See also :class:`~av.audio.plane.AudioPlane` and :class:`~av.video.plane.VideoPlane`.
    """

    def __cinit__(self, Frame frame, int index):
        self.frame = frame
        self.index = index

    def __repr__(self):
        return '<av.%s %d bytes; buffer_ptr=0x%x; at 0x%x>' % (
            self.__class__.__name__,
            self.buffer_size,
            self.buffer_ptr,
            id(self),
        )

    cdef void* _buffer_ptr(self):
        return self.frame.ptr.extended_data[self.index]



cimport libav as lib

from .option cimport Option, OptionChoice, wrap_option, wrap_option_choice


cdef object _cinit_sentinel = object()

cdef Descriptor wrap_avclass(const lib.AVClass *ptr):
    if ptr == NULL:
        return None
    cdef Descriptor obj = Descriptor(_cinit_sentinel)
    obj.ptr = ptr
    return obj


cdef class Descriptor(object):

    def __cinit__(self, sentinel):
        if sentinel is not _cinit_sentinel:
            raise RuntimeError('Cannot construct av.Descriptor')

    property name:
        def __get__(self): return self.ptr.class_name if self.ptr.class_name else None

    property options:
        def __get__(self):
            cdef const lib.AVOption *ptr = self.ptr.option
            cdef const lib.AVOption *choice_ptr
            cdef Option option
            cdef OptionChoice option_choice
            cdef bint choice_is_default
            if self._options is None:
                options = []
                ptr = self.ptr.option
                while ptr != NULL and ptr.name != NULL:
                    if ptr.type == lib.AV_OPT_TYPE_CONST:
                        ptr += 1
                        continue
                    choices = []
                    if ptr.unit != NULL:  # option has choices (matching const options)
                        choice_ptr = self.ptr.option
                        while choice_ptr != NULL and choice_ptr.name != NULL:
                            if choice_ptr.type != lib.AV_OPT_TYPE_CONST or choice_ptr.unit != ptr.unit:
                                choice_ptr += 1
                                continue
                            choice_is_default = (choice_ptr.default_val.i64 == ptr.default_val.i64 or
                                                 ptr.type == lib.AV_OPT_TYPE_FLAGS and
                                                 choice_ptr.default_val.i64 & ptr.default_val.i64)
                            option_choice = wrap_option_choice(choice_ptr, choice_is_default)
                            choices.append(option_choice)
                            choice_ptr += 1
                    option = wrap_option(tuple(choices), ptr)
                    options.append(option)
                    ptr += 1
                self._options = tuple(options)
            return self._options

    def __repr__(self):
        return '<%s %s at 0x%x>' % (self.__class__.__name__, self.name, id(self))



from fractions import Fraction
from unittest import SkipTest
import errno

import numpy as np

from av import AudioFrame, VideoFrame
from av.audio.frame import format_dtypes
from av.filter import Filter, Graph
import av

from .common import Image, TestCase, fate_suite


def generate_audio_frame(
    frame_num, input_format="s16", layout="stereo", sample_rate=44100, frame_size=1024
):
    """
    Generate audio frame representing part of the sinusoidal wave
    :param input_format: default: s16
    :param layout: default: stereo
    :param sample_rate: default: 44100
    :param frame_size: default: 1024
    :param frame_num: frame number
    :return: audio frame for sinusoidal wave audio signal slice
    """
    frame = AudioFrame(format=input_format, layout=layout, samples=frame_size)
    frame.sample_rate = sample_rate
    frame.pts = frame_num * frame_size

    for i in range(len(frame.layout.channels)):
        data = np.zeros(frame_size, dtype=format_dtypes[input_format])
        for j in range(frame_size):
            data[j] = np.sin(2 * np.pi * (frame_num + j) * (i + 1) / float(frame_size))
        frame.planes[i].update(data)

    return frame


def pull_until_blocked(graph):
    frames = []
    while True:
        try:
            frames.append(graph.pull())
        except av.utils.AVError as e:
            if e.errno != errno.EAGAIN:
                raise
            return frames


class TestFilters(TestCase):
    def test_filter_descriptor(self):

        f = Filter("testsrc")
        self.assertEqual(f.name, "testsrc")
        self.assertEqual(f.description, "Generate test pattern.")
        self.assertFalse(f.dynamic_inputs)
        self.assertEqual(len(f.inputs), 0)
        self.assertFalse(f.dynamic_outputs)
        self.assertEqual(len(f.outputs), 1)
        self.assertEqual(f.outputs[0].name, "default")
        self.assertEqual(f.outputs[0].type, "video")

    def test_dynamic_filter_descriptor(self):

        f = Filter("split")
        self.assertFalse(f.dynamic_inputs)
        self.assertEqual(len(f.inputs), 1)
        self.assertTrue(f.dynamic_outputs)
        self.assertEqual(len(f.outputs), 0)

    def test_generator_graph(self):

        graph = Graph()
        src = graph.add("testsrc")
        lutrgb = graph.add(
            "lutrgb",
            "r=maxval+minval-val:g=maxval+minval-val:b=maxval+minval-val",
            name="invert",
        )
        sink = graph.add("buffersink")
        src.link_to(lutrgb)
        lutrgb.link_to(sink)

        # pads and links
        self.assertIs(src.outputs[0].link.output, lutrgb.inputs[0])
        self.assertIs(lutrgb.inputs[0].link.input, src.outputs[0])

        frame = sink.pull()
        self.assertIsInstance(frame, VideoFrame)

        if Image:
            frame.to_image().save(self.sandboxed("mandelbrot2.png"))

    def test_auto_find_sink(self):

        graph = Graph()
        src = graph.add("testsrc")
        src.link_to(graph.add("buffersink"))
        graph.configure()

        frame = graph.pull()

        if Image:
            frame.to_image().save(self.sandboxed("mandelbrot3.png"))

    def test_delegate_sink(self):

        graph = Graph()
        src = graph.add("testsrc")
        src.link_to(graph.add("buffersink"))
        graph.configure()

        frame = src.pull()

        if Image:
            frame.to_image().save(self.sandboxed("mandelbrot4.png"))

    def test_haldclut_graph(self):

        raise SkipTest()

        graph = Graph()

        img = Image.open(fate_suite("png1/lena-rgb24.png"))
        frame = VideoFrame.from_image(img)
        img_source = graph.add_buffer(frame)

        hald_img = Image.open("hald_7.png")
        hald_frame = VideoFrame.from_image(hald_img)
        hald_source = graph.add_buffer(hald_frame)

        hald_filter = graph.add("haldclut")

        sink = graph.add("buffersink")

        img_source.link(0, hald_filter, 0)
        hald_source.link(0, hald_filter, 1)
        hald_filter.link(0, sink, 0)
        graph.config()

        self.assertIs(img_source.outputs[0].linked_to, hald_filter.inputs[0])
        self.assertIs(hald_source.outputs[0].linked_to, hald_filter.inputs[1])
        self.assertIs(hald_filter.outputs[0].linked_to, sink.inputs[0])

        hald_source.push(hald_frame)

        img_source.push(frame)

        frame = sink.pull()
        self.assertIsInstance(frame, VideoFrame)
        frame.to_image().save(self.sandboxed("filtered.png"))

    def test_audio_buffer_sink(self):
        graph = Graph()
        audio_buffer = graph.add_abuffer(
            format="fltp",
            sample_rate=48000,
            layout="stereo",
            time_base=Fraction(1, 48000),
        )
        audio_buffer.link_to(graph.add("abuffersink"))
        graph.configure()

        try:
            graph.pull()
        except OSError as e:
            # we haven't pushed any input so expect no frames / EAGAIN
            if e.errno != errno.EAGAIN:
                raise

    @staticmethod
    def link_nodes(*nodes):
        for c, n in zip(nodes, nodes[1:]):
            c.link_to(n)

    def test_audio_buffer_resample(self):
        graph = Graph()
        self.link_nodes(
            graph.add_abuffer(
                format="fltp",
                sample_rate=48000,
                layout="stereo",
                time_base=Fraction(1, 48000),
            ),
            graph.add(
                "aformat", "sample_fmts=s16:sample_rates=44100:channel_layouts=stereo"
            ),
            graph.add("abuffersink"),
        )
        graph.configure()

        graph.push(
            generate_audio_frame(
                0, input_format="fltp", layout="stereo", sample_rate=48000
            )
        )
        out_frame = graph.pull()
        self.assertEqual(out_frame.format.name, "s16")
        self.assertEqual(out_frame.layout.name, "stereo")
        self.assertEqual(out_frame.sample_rate, 44100)

    def test_audio_buffer_volume_filter(self):
        graph = Graph()
        self.link_nodes(
            graph.add_abuffer(
                format="fltp",
                sample_rate=48000,
                layout="stereo",
                time_base=Fraction(1, 48000),
            ),
            graph.add("volume", volume="0.5"),
            graph.add("abuffersink"),
        )
        graph.configure()

        input_frame = generate_audio_frame(
            0, input_format="fltp", layout="stereo", sample_rate=48000
        )
        graph.push(input_frame)

        out_frame = graph.pull()
        self.assertEqual(out_frame.format.name, "fltp")
        self.assertEqual(out_frame.layout.name, "stereo")
        self.assertEqual(out_frame.sample_rate, 48000)

        input_data = input_frame.to_ndarray()
        output_data = out_frame.to_ndarray()

        self.assertTrue(
            np.allclose(input_data * 0.5, output_data), "Check that volume is reduced"
        )

    def test_video_buffer(self):
        input_container = av.open(format="lavfi", file="color=c=pink:duration=1:r=30")
        input_video_stream = input_container.streams.video[0]

        graph = av.filter.Graph()
        buffer = graph.add_buffer(template=input_video_stream)
        bwdif = graph.add("bwdif", "send_field:tff:all")
        buffersink = graph.add("buffersink")
        buffer.link_to(bwdif)
        bwdif.link_to(buffersink)
        graph.configure()

        for frame in input_container.decode():
            self.assertEqual(frame.time_base, Fraction(1, 30))
            graph.push(frame)
            filtered_frames = pull_until_blocked(graph)

            if frame.pts == 0:
                # no output for the first input frame
                self.assertEqual(len(filtered_frames), 0)
            else:
                # we expect two filtered frames per input frame
                self.assertEqual(len(filtered_frames), 2)

                self.assertEqual(filtered_frames[0].pts, (frame.pts - 1) * 2)
                self.assertEqual(filtered_frames[0].time_base, Fraction(1, 60))

                self.assertEqual(filtered_frames[1].pts, (frame.pts - 1) * 2 + 1)
                self.assertEqual(filtered_frames[1].time_base, Fraction(1, 60))

    def test_EOF(self):
        input_container = av.open(format="lavfi", file="color=c=pink:duration=1:r=30")
        video_stream = input_container.streams.video[0]

        graph = av.filter.Graph()
        video_in = graph.add_buffer(template=video_stream)
        palette_gen_filter = graph.add("palettegen")
        video_out = graph.add("buffersink")
        video_in.link_to(palette_gen_filter)
        palette_gen_filter.link_to(video_out)
        graph.configure()

        for frame in input_container.decode(video=0):
            graph.push(frame)

        graph.push(None)

        # if we do not push None, we get a BlockingIOError
        palette_frame = graph.pull()

        self.assertIsInstance(palette_frame, av.VideoFrame)
        self.assertEqual(palette_frame.width, 16)
        self.assertEqual(palette_frame.height, 16)



import errno
import traceback

import av

from .common import TestCase, is_windows


class TestErrorBasics(TestCase):
    def test_stringify(self):

        for cls in (av.ValueError, av.FileNotFoundError, av.DecoderNotFoundError):
            e = cls(1, "foo")
            self.assertEqual(str(e), "[Errno 1] foo")
            self.assertEqual(repr(e), "{}(1, 'foo')".format(cls.__name__))
            self.assertEqual(
                traceback.format_exception_only(cls, e)[-1],
                "{}{}: [Errno 1] foo\n".format(
                    "av.error.",
                    cls.__name__,
                ),
            )

        for cls in (av.ValueError, av.FileNotFoundError, av.DecoderNotFoundError):
            e = cls(1, "foo", "bar.txt")
            self.assertEqual(str(e), "[Errno 1] foo: 'bar.txt'")
            self.assertEqual(repr(e), "{}(1, 'foo', 'bar.txt')".format(cls.__name__))
            self.assertEqual(
                traceback.format_exception_only(cls, e)[-1],
                "{}{}: [Errno 1] foo: 'bar.txt'\n".format(
                    "av.error.",
                    cls.__name__,
                ),
            )

    def test_bases(self):

        self.assertTrue(issubclass(av.ValueError, ValueError))
        self.assertTrue(issubclass(av.ValueError, av.FFmpegError))

        self.assertTrue(issubclass(av.FileNotFoundError, FileNotFoundError))
        self.assertTrue(issubclass(av.FileNotFoundError, OSError))
        self.assertTrue(issubclass(av.FileNotFoundError, av.FFmpegError))

    def test_filenotfound(self):
        """Catch using builtin class on Python 3.3"""
        try:
            av.open("does not exist")
        except FileNotFoundError as e:
            self.assertEqual(e.errno, errno.ENOENT)
            if is_windows:
                self.assertTrue(
                    e.strerror
                    in ["Error number -2 occurred", "No such file or directory"]
                )
            else:
                self.assertEqual(e.strerror, "No such file or directory")
            self.assertEqual(e.filename, "does not exist")
        else:
            self.fail("no exception raised")

    def test_buffertoosmall(self):
        """Throw an exception from an enum."""
        try:
            av.error.err_check(-av.error.BUFFER_TOO_SMALL.value)
        except av.BufferTooSmallError as e:
            self.assertEqual(e.errno, av.error.BUFFER_TOO_SMALL.value)
        else:
            self.fail("no exception raised")



import warnings

from av import deprecation

from .common import TestCase


class TestDeprecations(TestCase):
    def test_method(self):
        class Example(object):
            def __init__(self, x=100):
                self.x = x

            @deprecation.method
            def foo(self, a, b):
                return self.x + a + b

        obj = Example()

        with warnings.catch_warnings(record=True) as captured:
            self.assertEqual(obj.foo(20, b=3), 123)
            self.assertIn("Example.foo is deprecated", captured[0].message.args[0])

    def test_renamed_attr(self):
        class Example(object):

            new_value = "foo"
            old_value = deprecation.renamed_attr("new_value")

            def new_func(self, a, b):
                return a + b

            old_func = deprecation.renamed_attr("new_func")

        obj = Example()

        with warnings.catch_warnings(record=True) as captured:

            self.assertEqual(obj.old_value, "foo")
            self.assertIn(
                "Example.old_value is deprecated", captured[0].message.args[0]
            )

            obj.old_value = "bar"
            self.assertIn(
                "Example.old_value is deprecated", captured[1].message.args[0]
            )

        with warnings.catch_warnings(record=True) as captured:
            self.assertEqual(obj.old_func(1, 2), 3)
            self.assertIn("Example.old_func is deprecated", captured[0].message.args[0])



cdef extern from "libavutil/channel_layout.h" nogil:

    # This is not a comprehensive list.
    cdef uint64_t AV_CH_LAYOUT_MONO
    cdef uint64_t AV_CH_LAYOUT_STEREO
    cdef uint64_t AV_CH_LAYOUT_2POINT1
    cdef uint64_t AV_CH_LAYOUT_4POINT0
    cdef uint64_t AV_CH_LAYOUT_5POINT0_BACK
    cdef uint64_t AV_CH_LAYOUT_5POINT1_BACK
    cdef uint64_t AV_CH_LAYOUT_6POINT1
    cdef uint64_t AV_CH_LAYOUT_7POINT1



from libc.stdint cimport int64_t
cimport libav as lib

from av.bytesource cimport ByteSource
from av.codec.codec cimport Codec
from av.frame cimport Frame
from av.packet cimport Packet


cdef class CodecContext(object):

    cdef lib.AVCodecContext *ptr

    # Whether AVCodecContext.extradata should be de-allocated upon destruction.
    cdef bint extradata_set

    # Used as a signal that this is within a stream, and also for us to access
    # that stream. This is set "manually" by the stream after constructing
    # this object.
    cdef int stream_index

    cdef lib.AVCodecParserContext *parser

    cdef _init(self, lib.AVCodecContext *ptr, const lib.AVCodec *codec)

    cdef readonly Codec codec

    cdef public dict options

    # Public API.
    cpdef open(self, bint strict=?)
    cpdef close(self, bint strict=?)

    cdef _set_default_time_base(self)

    # Wraps both versions of the transcode API, returning lists.
    cpdef encode(self, Frame frame=?)
    cpdef decode(self, Packet packet=?)

    # Used by both transcode APIs to setup user-land objects.
    # TODO: Remove the `Packet` from `_setup_decoded_frame` (because flushing
    # packets are bogus). It should take all info it needs from the context and/or stream.
    cdef _prepare_frames_for_encode(self, Frame frame)
    cdef _setup_encoded_packet(self, Packet)
    cdef _setup_decoded_frame(self, Frame, Packet)

    # Implemented by base for the generic send/recv API.
    # Note that the user cannot send without recieving. This is because
    # _prepare_frames_for_encode may expand a frame into multiple (e.g. when
    # resampling audio to a higher rate but with fixed size frames), and the
    # send/recv buffer may be limited to a single frame. Ergo, we need to flush
    # the buffer as often as possible.
    cdef _send_frame_and_recv(self, Frame frame)
    cdef _recv_packet(self)
    cdef _send_packet_and_recv(self, Packet packet)
    cdef _recv_frame(self)

    # Implemented by children for the generic send/recv API, so we have the
    # correct subclass of Frame.
    cdef Frame _next_frame
    cdef Frame _alloc_next_frame(self)


cdef CodecContext wrap_codec_context(lib.AVCodecContext*, const lib.AVCodec*)



from av import ContainerFormat
from av.option import Option, OptionType

from .common import TestCase


class TestOptions(TestCase):
    def test_mov_options(self):

        mov = ContainerFormat("mov")
        options = mov.descriptor.options
        by_name = {opt.name: opt for opt in options}

        opt = by_name.get("use_absolute_path")

        self.assertIsInstance(opt, Option)
        self.assertEqual(opt.name, "use_absolute_path")

        # This was not a good option to actually test.
        self.assertIn(opt.type, (OptionType.BOOL, OptionType.INT))



cimport libav as lib

from av.bytesource cimport bytesource
from av.error cimport err_check
from av.utils cimport avrational_to_fraction, to_avrational

from av import deprecation


cdef class Packet(Buffer):

    """A packet of encoded data within a :class:`~av.format.Stream`.

    This may, or may not include a complete object within a stream.
    :meth:`decode` must be called to extract encoded data.

    """

    def __cinit__(self, input=None):
        with nogil:
            self.ptr = lib.av_packet_alloc()

    def __init__(self, input=None):

        cdef size_t size = 0
        cdef ByteSource source = None

        if input is None:
            return

        if isinstance(input, (int, long)):
            size = input
        else:
            source = bytesource(input)
            size = source.length

        if size:
            err_check(lib.av_new_packet(self.ptr, size))

        if source is not None:
            self.update(source)
            # TODO: Hold onto the source, and copy its pointer
            # instead of its data.
            # self.source = source

    def __dealloc__(self):
        with nogil:
            lib.av_packet_free(&self.ptr)

    def __repr__(self):
        return '<av.%s of #%d, dts=%s, pts=%s; %s bytes at 0x%x>' % (
            self.__class__.__name__,
            self._stream.index if self._stream else 0,
            self.dts,
            self.pts,
            self.ptr.size,
            id(self),
        )

    # Buffer protocol.
    cdef size_t _buffer_size(self):
        return self.ptr.size
    cdef void* _buffer_ptr(self):
        return self.ptr.data

    cdef _rebase_time(self, lib.AVRational dst):

        if not dst.num:
            raise ValueError('Cannot rebase to zero time.')

        if not self._time_base.num:
            self._time_base = dst
            return

        if self._time_base.num == dst.num and self._time_base.den == dst.den:
            return

        lib.av_packet_rescale_ts(self.ptr, self._time_base, dst)

        self._time_base = dst

    def decode(self):
        """
        Send the packet's data to the decoder and return a list of
        :class:`.AudioFrame`, :class:`.VideoFrame` or :class:`.SubtitleSet`.
        """
        return self._stream.decode(self)

    @deprecation.method
    def decode_one(self):
        """
        Send the packet's data to the decoder and return the first decoded frame.

        Returns ``None`` if there is no frame.

        .. warning:: This method is deprecated, as it silently discards any
                     other frames which were decoded.
        """
        res = self._stream.decode(self)
        return res[0] if res else None

    property stream_index:
        def __get__(self):
            return self.ptr.stream_index

    property stream:
        """
        The :class:`Stream` this packet was demuxed from.
        """
        def __get__(self):
            return self._stream

        def __set__(self, Stream stream):
            self._stream = stream
            self.ptr.stream_index = stream.ptr.index

    property time_base:
        """
        The unit of time (in fractional seconds) in which timestamps are expressed.

        :type: fractions.Fraction
        """
        def __get__(self):
            return avrational_to_fraction(&self._time_base)

        def __set__(self, value):
            to_avrational(value, &self._time_base)

    property pts:
        """
        The presentation timestamp in :attr:`time_base` units for this packet.

        This is the time at which the packet should be shown to the user.

        :type: int
        """
        def __get__(self):
            if self.ptr.pts != lib.AV_NOPTS_VALUE:
                return self.ptr.pts

        def __set__(self, v):
            if v is None:
                self.ptr.pts = lib.AV_NOPTS_VALUE
            else:
                self.ptr.pts = v

    property dts:
        """
        The decoding timestamp in :attr:`time_base` units for this packet.

        :type: int
        """
        def __get__(self):
            if self.ptr.dts != lib.AV_NOPTS_VALUE:
                return self.ptr.dts

        def __set__(self, v):
            if v is None:
                self.ptr.dts = lib.AV_NOPTS_VALUE
            else:
                self.ptr.dts = v

    property pos:
        """
        The byte position of this packet within the :class:`.Stream`.

        Returns `None` if it is not known.

        :type: int
        """
        def __get__(self):
            if self.ptr.pos != -1:
                return self.ptr.pos

    property size:
        """
        The size in bytes of this packet's data.

        :type: int
        """
        def __get__(self):
            return self.ptr.size

    property duration:
        """
        The duration in :attr:`time_base` units for this packet.

        Returns `None` if it is not known.

        :type: int
        """
        def __get__(self):
            if self.ptr.duration != lib.AV_NOPTS_VALUE:
                return self.ptr.duration

    property is_keyframe:
        def __get__(self): return bool(self.ptr.flags & lib.AV_PKT_FLAG_KEY)

    property is_corrupt:
        def __get__(self): return bool(self.ptr.flags & lib.AV_PKT_FLAG_CORRUPT)



cimport libav as lib


cdef class AudioFormat(object):

    cdef lib.AVSampleFormat sample_fmt

    cdef _init(self, lib.AVSampleFormat sample_fmt)


cdef AudioFormat get_audio_format(lib.AVSampleFormat format)



cimport libav as lib

from av.video.frame cimport VideoFrame


cdef class VideoReformatter(object):

    cdef lib.SwsContext *ptr

    cdef _reformat(self, VideoFrame frame, int width, int height,
                   lib.AVPixelFormat format, int src_colorspace,
                   int dst_colorspace, int interpolation)



from libc.stdint cimport int64_t, uint8_t, uint64_t

from fractions import Fraction

cimport libav as lib

from av.error cimport err_check


# === DICTIONARIES ===
# ====================

cdef _decode(char *s, encoding, errors):
    return (<bytes>s).decode(encoding, errors)

cdef bytes _encode(s, encoding, errors):
    return s.encode(encoding, errors)

cdef dict avdict_to_dict(lib.AVDictionary *input, str encoding, str errors):
    cdef lib.AVDictionaryEntry *element = NULL
    cdef dict output = {}
    while True:
        element = lib.av_dict_get(input, "", element, lib.AV_DICT_IGNORE_SUFFIX)
        if element == NULL:
            break
        output[_decode(element.key, encoding, errors)] = _decode(element.value, encoding, errors)
    return output


cdef dict_to_avdict(lib.AVDictionary **dst, dict src, str encoding, str errors):
    lib.av_dict_free(dst)
    for key, value in src.items():
        err_check(lib.av_dict_set(dst, _encode(key, encoding, errors),
                                  _encode(value, encoding, errors), 0))


# === FRACTIONS ===
# =================

cdef object avrational_to_fraction(const lib.AVRational *input):
    if input.num and input.den:
        return Fraction(input.num, input.den)


cdef object to_avrational(object value, lib.AVRational *input):

    if value is None:
        input.num = 0
        input.den = 1
        return

    if isinstance(value, Fraction):
        frac = value
    else:
        frac = Fraction(value)

    input.num = frac.numerator
    input.den = frac.denominator


# === OTHER ===
# =============


cdef check_ndarray(object array, object dtype, int ndim):
    """
    Check a numpy array has the expected data type and number of dimensions.
    """
    if array.dtype != dtype:
        raise ValueError(f"Expected numpy array with dtype `{dtype}` but got `{array.dtype}`")
    if array.ndim != ndim:
        raise ValueError(f"Expected numpy array with ndim `{ndim}` but got `{array.ndim}`")


cdef check_ndarray_shape(object array, bint ok):
    """
    Check a numpy array has the expected shape.
    """
    if not ok:
        raise ValueError(f"Unexpected numpy array shape `{array.shape}`")


cdef flag_in_bitfield(uint64_t bitfield, uint64_t flag):
    # Not every flag exists in every version of FFMpeg, so we define them to 0.
    if not flag:
        return None
    return bool(bitfield & flag)


# === BACKWARDS COMPAT ===

from .error import FFmpegError as AVError
from .error import err_check



import argparse


def main():

    parser = argparse.ArgumentParser()
    parser.add_argument("--codecs", action="store_true")
    parser.add_argument("--version", action="store_true")
    args = parser.parse_args()

    # ---

    if args.version:

        import av
        import av._core

        print("PyAV v" + av.__version__)

        by_config = {}
        for libname, config in sorted(av._core.library_meta.items()):
            version = config["version"]
            if version[0] >= 0:
                by_config.setdefault(
                    (config["configuration"], config["license"]), []
                ).append((libname, config))
        for (config, license), libs in sorted(by_config.items()):
            print("library configuration:", config)
            print("library license:", license)
            for libname, config in libs:
                version = config["version"]
                print(
                    "%-13s %3d.%3d.%3d" % (libname, version[0], version[1], version[2])
                )

    if args.codecs:
        from av.codec.codec import dump_codecs

        dump_codecs()


if __name__ == "__main__":
    main()



from PIL import Image
import numpy as np

import av
import av.datasets


container = av.open(
    av.datasets.curated("pexels/time-lapse-video-of-sunset-by-the-sea-854400.mp4")
)
container.streams.video[0].thread_type = "AUTO"  # Go faster!

columns = []
for frame in container.decode(video=0):

    print(frame)
    array = frame.to_ndarray(format="rgb24")

    # Collapse down to a column.
    column = array.mean(axis=1)

    # Convert to bytes, as the `mean` turned our array into floats.
    column = column.clip(0, 255).astype("uint8")

    # Get us in the right shape for the `hstack` below.
    column = column.reshape(-1, 1, 3)

    columns.append(column)

# Close the file, free memory
container.close()

full_array = np.hstack(columns)
full_img = Image.fromarray(full_array, "RGB")
full_img = full_img.resize((800, 200))
full_img.save("barcode.jpg", quality=85)



from fractions import Fraction

import av

from .common import TestCase, fate_suite


class TestDecode(TestCase):
    def test_decoded_video_frame_count(self):

        container = av.open(fate_suite("h264/interlaced_crop.mp4"))
        video_stream = next(s for s in container.streams if s.type == "video")

        self.assertIs(video_stream, container.streams.video[0])

        frame_count = 0

        for packet in container.demux(video_stream):
            for frame in packet.decode():
                frame_count += 1

        self.assertEqual(frame_count, video_stream.frames)

    def test_decode_audio_corrupt(self):
        # write an empty file
        path = self.sandboxed("empty.flac")
        with open(path, "wb"):
            pass

        packet_count = 0
        frame_count = 0

        with av.open(path) as container:
            for packet in container.demux(audio=0):
                for frame in packet.decode():
                    frame_count += 1
                packet_count += 1

        self.assertEqual(packet_count, 1)
        self.assertEqual(frame_count, 0)

    def test_decode_audio_sample_count(self):

        container = av.open(fate_suite("audio-reference/chorusnoise_2ch_44kHz_s16.wav"))
        audio_stream = next(s for s in container.streams if s.type == "audio")

        self.assertIs(audio_stream, container.streams.audio[0])

        sample_count = 0

        for packet in container.demux(audio_stream):
            for frame in packet.decode():
                sample_count += frame.samples

        total_samples = (
            audio_stream.duration * audio_stream.sample_rate.numerator
        ) / audio_stream.time_base.denominator
        self.assertEqual(sample_count, total_samples)

    def test_decoded_time_base(self):

        container = av.open(fate_suite("h264/interlaced_crop.mp4"))
        stream = container.streams.video[0]

        self.assertEqual(stream.time_base, Fraction(1, 25))

        for packet in container.demux(stream):
            for frame in packet.decode():
                self.assertEqual(packet.time_base, frame.time_base)
                self.assertEqual(stream.time_base, frame.time_base)
                return

    def test_decoded_motion_vectors(self):

        container = av.open(fate_suite("h264/interlaced_crop.mp4"))
        stream = container.streams.video[0]
        codec_context = stream.codec_context
        codec_context.options = {"flags2": "+export_mvs"}

        for packet in container.demux(stream):
            for frame in packet.decode():
                vectors = frame.side_data.get("MOTION_VECTORS")
                if frame.key_frame:
                    # Key frame don't have motion vectors
                    assert vectors is None
                else:
                    assert len(vectors) > 0
                    return

    def test_decoded_motion_vectors_no_flag(self):

        container = av.open(fate_suite("h264/interlaced_crop.mp4"))
        stream = container.streams.video[0]

        for packet in container.demux(stream):
            for frame in packet.decode():
                vectors = frame.side_data.get("MOTION_VECTORS")
                if not frame.key_frame:
                    assert vectors is None
                    return

    def test_decode_video_corrupt(self):
        # write an empty file
        path = self.sandboxed("empty.h264")
        with open(path, "wb"):
            pass

        packet_count = 0
        frame_count = 0

        with av.open(path) as container:
            for packet in container.demux(video=0):
                for frame in packet.decode():
                    frame_count += 1
                packet_count += 1

        self.assertEqual(packet_count, 1)
        self.assertEqual(frame_count, 0)



cimport libav as lib

from av.enum cimport define_enum
from av.utils cimport flag_in_bitfield


cdef object _cinit_sentinel = object()

cdef Option wrap_option(tuple choices, const lib.AVOption *ptr):
    if ptr == NULL:
        return None
    cdef Option obj = Option(_cinit_sentinel)
    obj.ptr = ptr
    obj.choices = choices
    return obj


OptionType = define_enum('OptionType', __name__, (
    ('FLAGS', lib.AV_OPT_TYPE_FLAGS),
    ('INT', lib.AV_OPT_TYPE_INT),
    ('INT64', lib.AV_OPT_TYPE_INT64),
    ('DOUBLE', lib.AV_OPT_TYPE_DOUBLE),
    ('FLOAT', lib.AV_OPT_TYPE_FLOAT),
    ('STRING', lib.AV_OPT_TYPE_STRING),
    ('RATIONAL', lib.AV_OPT_TYPE_RATIONAL),
    ('BINARY', lib.AV_OPT_TYPE_BINARY),
    ('DICT', lib.AV_OPT_TYPE_DICT),
    # ('UINT64', lib.AV_OPT_TYPE_UINT64), # Added recently, and not yet used AFAICT.
    ('CONST', lib.AV_OPT_TYPE_CONST),
    ('IMAGE_SIZE', lib.AV_OPT_TYPE_IMAGE_SIZE),
    ('PIXEL_FMT', lib.AV_OPT_TYPE_PIXEL_FMT),
    ('SAMPLE_FMT', lib.AV_OPT_TYPE_SAMPLE_FMT),
    ('VIDEO_RATE', lib.AV_OPT_TYPE_VIDEO_RATE),
    ('DURATION', lib.AV_OPT_TYPE_DURATION),
    ('COLOR', lib.AV_OPT_TYPE_COLOR),
    ('CHANNEL_LAYOUT', lib.AV_OPT_TYPE_CHANNEL_LAYOUT),
    ('BOOL', lib.AV_OPT_TYPE_BOOL),
))

cdef tuple _INT_TYPES = (
    lib.AV_OPT_TYPE_FLAGS,
    lib.AV_OPT_TYPE_INT,
    lib.AV_OPT_TYPE_INT64,
    lib.AV_OPT_TYPE_PIXEL_FMT,
    lib.AV_OPT_TYPE_SAMPLE_FMT,
    lib.AV_OPT_TYPE_DURATION,
    lib.AV_OPT_TYPE_CHANNEL_LAYOUT,
    lib.AV_OPT_TYPE_BOOL,
)

OptionFlags = define_enum('OptionFlags', __name__, (
    ('ENCODING_PARAM', lib.AV_OPT_FLAG_ENCODING_PARAM),
    ('DECODING_PARAM', lib.AV_OPT_FLAG_DECODING_PARAM),
    ('AUDIO_PARAM', lib.AV_OPT_FLAG_AUDIO_PARAM),
    ('VIDEO_PARAM', lib.AV_OPT_FLAG_VIDEO_PARAM),
    ('SUBTITLE_PARAM', lib.AV_OPT_FLAG_SUBTITLE_PARAM),
    ('EXPORT', lib.AV_OPT_FLAG_EXPORT),
    ('READONLY', lib.AV_OPT_FLAG_READONLY),
    ('FILTERING_PARAM', lib.AV_OPT_FLAG_FILTERING_PARAM),
), is_flags=True)

cdef class BaseOption(object):

    def __cinit__(self, sentinel):
        if sentinel is not _cinit_sentinel:
            raise RuntimeError('Cannot construct av.%s' % self.__class__.__name__)

    property name:
        def __get__(self):
            return self.ptr.name

    property help:
        def __get__(self):
            return self.ptr.help if self.ptr.help != NULL else ''

    property flags:
        def __get__(self):
            return self.ptr.flags

    # Option flags
    property is_encoding_param:
        def __get__(self):
            return flag_in_bitfield(self.ptr.flags, lib.AV_OPT_FLAG_ENCODING_PARAM)
    property is_decoding_param:
        def __get__(self):
            return flag_in_bitfield(self.ptr.flags, lib.AV_OPT_FLAG_DECODING_PARAM)
    property is_audio_param:
        def __get__(self):
            return flag_in_bitfield(self.ptr.flags, lib.AV_OPT_FLAG_AUDIO_PARAM)
    property is_video_param:
        def __get__(self):
            return flag_in_bitfield(self.ptr.flags, lib.AV_OPT_FLAG_VIDEO_PARAM)
    property is_subtitle_param:
        def __get__(self):
            return flag_in_bitfield(self.ptr.flags, lib.AV_OPT_FLAG_SUBTITLE_PARAM)
    property is_export:
        def __get__(self):
            return flag_in_bitfield(self.ptr.flags, lib.AV_OPT_FLAG_EXPORT)
    property is_readonly:
        def __get__(self):
            return flag_in_bitfield(self.ptr.flags, lib.AV_OPT_FLAG_READONLY)
    property is_filtering_param:
        def __get__(self):
            return flag_in_bitfield(self.ptr.flags, lib.AV_OPT_FLAG_FILTERING_PARAM)


cdef class Option(BaseOption):

    property type:
        def __get__(self):
            return OptionType._get(self.ptr.type, create=True)

    property offset:
        """
        This can be used to find aliases of an option.
        Options in a particular descriptor with the same offset are aliases.
        """
        def __get__(self):
            return self.ptr.offset

    property default:
        def __get__(self):
            if self.ptr.type in _INT_TYPES:
                return self.ptr.default_val.i64
            if self.ptr.type in (lib.AV_OPT_TYPE_DOUBLE, lib.AV_OPT_TYPE_FLOAT,
                                 lib.AV_OPT_TYPE_RATIONAL):
                return self.ptr.default_val.dbl
            if self.ptr.type in (lib.AV_OPT_TYPE_STRING, lib.AV_OPT_TYPE_BINARY,
                                 lib.AV_OPT_TYPE_IMAGE_SIZE, lib.AV_OPT_TYPE_VIDEO_RATE,
                                 lib.AV_OPT_TYPE_COLOR):
                return self.ptr.default_val.str if self.ptr.default_val.str != NULL else ''

    def _norm_range(self, value):
        if self.ptr.type in _INT_TYPES:
            return int(value)
        return value

    property min:
        def __get__(self):
            return self._norm_range(self.ptr.min)

    property max:
        def __get__(self):
            return self._norm_range(self.ptr.max)

    def __repr__(self):
        return '<av.%s %s (%s at *0x%x) at 0x%x>' % (
            self.__class__.__name__,
            self.name,
            self.type,
            self.offset,
            id(self),
        )


cdef OptionChoice wrap_option_choice(const lib.AVOption *ptr, bint is_default):
    if ptr == NULL:
        return None
    cdef OptionChoice obj = OptionChoice(_cinit_sentinel)
    obj.ptr = ptr
    obj.is_default = is_default
    return obj


cdef class OptionChoice(BaseOption):
    """
    Represents AV_OPT_TYPE_CONST options which are essentially
    choices of non-const option with same unit.
    """

    property value:
        def __get__(self):
            return self.ptr.default_val.i64

    def __repr__(self):
        return '<av.%s %s at 0x%x>' % (self.__class__.__name__, self.name, id(self))



cimport libav as lib


# Initialise libraries.
lib.avformat_network_init()
lib.avdevice_register_all()

# Exports.
time_base = lib.AV_TIME_BASE


cdef decode_version(v):
    if v < 0:
        return (-1, -1, -1)
    cdef int major = (v >> 16) & 0xff
    cdef int minor = (v >> 8) & 0xff
    cdef int micro = (v) & 0xff
    return (major, minor, micro)

library_meta = {
    'libavutil': dict(
        version=decode_version(lib.avutil_version()),
        configuration=lib.avutil_configuration(),
        license=lib.avutil_license()
    ),
    'libavcodec': dict(
        version=decode_version(lib.avcodec_version()),
        configuration=lib.avcodec_configuration(),
        license=lib.avcodec_license()
    ),
    'libavformat': dict(
        version=decode_version(lib.avformat_version()),
        configuration=lib.avformat_configuration(),
        license=lib.avformat_license()
    ),
    'libavdevice': dict(
        version=decode_version(lib.avdevice_version()),
        configuration=lib.avdevice_configuration(),
        license=lib.avdevice_license()
    ),
    'libavfilter': dict(
        version=decode_version(lib.avfilter_version()),
        configuration=lib.avfilter_configuration(),
        license=lib.avfilter_license()
    ),
    'libswscale': dict(
        version=decode_version(lib.swscale_version()),
        configuration=lib.swscale_configuration(),
        license=lib.swscale_license()
    ),
    'libswresample': dict(
        version=decode_version(lib.swresample_version()),
        configuration=lib.swresample_configuration(),
        license=lib.swresample_license()
    ),
}

library_versions = {name: meta['version'] for name, meta in library_meta.items()}



cimport libav as lib

from av.packet cimport Packet


cdef class SubtitleProxy(object):

    cdef lib.AVSubtitle struct


cdef class SubtitleSet(object):

    cdef readonly Packet packet
    cdef SubtitleProxy proxy
    cdef readonly tuple rects


cdef class Subtitle(object):

    cdef SubtitleProxy proxy
    cdef lib.AVSubtitleRect *ptr
    cdef readonly bytes type

cdef class TextSubtitle(Subtitle):
    pass

cdef class ASSSubtitle(Subtitle):
    pass

cdef class BitmapSubtitle(Subtitle):

    cdef readonly planes

cdef class BitmapSubtitlePlane(object):

    cdef readonly BitmapSubtitle subtitle
    cdef readonly int index
    cdef readonly long buffer_size
    cdef void *_buffer



from unittest import SkipTest

import numpy

from av import VideoFrame

from .common import Image, TestCase, fate_png


class TestVideoFrameConstructors(TestCase):
    def test_invalid_pixel_format(self):
        with self.assertRaises(ValueError) as cm:
            VideoFrame(640, 480, "__unknown_pix_fmt")
        self.assertEqual(str(cm.exception), "not a pixel format: '__unknown_pix_fmt'")

    def test_null_constructor(self):
        frame = VideoFrame()
        self.assertEqual(frame.width, 0)
        self.assertEqual(frame.height, 0)
        self.assertEqual(frame.format.name, "yuv420p")

    def test_manual_yuv_constructor(self):
        frame = VideoFrame(640, 480, "yuv420p")
        self.assertEqual(frame.width, 640)
        self.assertEqual(frame.height, 480)
        self.assertEqual(frame.format.name, "yuv420p")

    def test_manual_rgb_constructor(self):
        frame = VideoFrame(640, 480, "rgb24")
        self.assertEqual(frame.width, 640)
        self.assertEqual(frame.height, 480)
        self.assertEqual(frame.format.name, "rgb24")


class TestVideoFramePlanes(TestCase):
    def test_null_planes(self):
        frame = VideoFrame()  # yuv420p
        self.assertEqual(len(frame.planes), 0)

    def test_yuv420p_planes(self):
        frame = VideoFrame(640, 480, "yuv420p")
        self.assertEqual(len(frame.planes), 3)
        self.assertEqual(frame.planes[0].width, 640)
        self.assertEqual(frame.planes[0].height, 480)
        self.assertEqual(frame.planes[0].line_size, 640)
        self.assertEqual(frame.planes[0].buffer_size, 640 * 480)
        for i in range(1, 3):
            self.assertEqual(frame.planes[i].width, 320)
            self.assertEqual(frame.planes[i].height, 240)
            self.assertEqual(frame.planes[i].line_size, 320)
            self.assertEqual(frame.planes[i].buffer_size, 320 * 240)

    def test_yuv420p_planes_align(self):
        # If we request 8-byte alignment for a width which is not a multiple of 8,
        # the line sizes are larger than the plane width.
        frame = VideoFrame(318, 238, "yuv420p")
        self.assertEqual(len(frame.planes), 3)
        self.assertEqual(frame.planes[0].width, 318)
        self.assertEqual(frame.planes[0].height, 238)
        self.assertEqual(frame.planes[0].line_size, 320)
        self.assertEqual(frame.planes[0].buffer_size, 320 * 238)
        for i in range(1, 3):
            self.assertEqual(frame.planes[i].width, 159)
            self.assertEqual(frame.planes[i].height, 119)
            self.assertEqual(frame.planes[i].line_size, 160)
            self.assertEqual(frame.planes[i].buffer_size, 160 * 119)

    def test_rgb24_planes(self):
        frame = VideoFrame(640, 480, "rgb24")
        self.assertEqual(len(frame.planes), 1)
        self.assertEqual(frame.planes[0].width, 640)
        self.assertEqual(frame.planes[0].height, 480)
        self.assertEqual(frame.planes[0].line_size, 640 * 3)
        self.assertEqual(frame.planes[0].buffer_size, 640 * 480 * 3)


class TestVideoFrameBuffers(TestCase):
    def test_memoryview_read(self):
        frame = VideoFrame(640, 480, "rgb24")
        frame.planes[0].update(b"01234" + (b"x" * (640 * 480 * 3 - 5)))
        mem = memoryview(frame.planes[0])
        self.assertEqual(mem.ndim, 1)
        self.assertEqual(mem.shape, (640 * 480 * 3,))
        self.assertFalse(mem.readonly)
        self.assertEqual(mem[1], 49)
        self.assertEqual(mem[:7], b"01234xx")
        mem[1] = 46
        self.assertEqual(mem[:7], b"0.234xx")


class TestVideoFrameImage(TestCase):
    def setUp(self):
        if not Image:
            raise SkipTest()

    def test_roundtrip(self):
        image = Image.open(fate_png())
        frame = VideoFrame.from_image(image)
        img = frame.to_image()
        img.save(self.sandboxed("roundtrip-high.jpg"))
        self.assertImagesAlmostEqual(image, img)

    def test_to_image_rgb24(self):
        sizes = [
            (318, 238),
            (320, 240),
            (500, 500),
        ]
        for width, height in sizes:
            frame = VideoFrame(width, height, format="rgb24")

            # fill video frame data
            for plane in frame.planes:
                ba = bytearray(plane.buffer_size)
                pos = 0
                for row in range(height):
                    for i in range(plane.line_size):
                        ba[pos] = i % 256
                        pos += 1
                plane.update(ba)

            # construct expected image data
            expected = bytearray(height * width * 3)
            pos = 0
            for row in range(height):
                for i in range(width * 3):
                    expected[pos] = i % 256
                    pos += 1

            img = frame.to_image()
            self.assertEqual(img.size, (width, height))
            self.assertEqual(img.tobytes(), expected)

    def test_to_image_with_dimensions(self):
        frame = VideoFrame(640, 480, format="rgb24")

        img = frame.to_image(width=320, height=240)
        self.assertEqual(img.size, (320, 240))


class TestVideoFrameNdarray(TestCase):
    def assertPixelValue16(self, plane, expected, byteorder: str):
        view = memoryview(plane)
        if byteorder == "big":
            self.assertEqual(view[0], (expected >> 8) & 0xFF)
            self.assertEqual(view[1], expected & 0xFF)
        else:
            self.assertEqual(view[0], expected & 0xFF)
            self.assertEqual(view[1], (expected >> 8) & 0xFF)

    def test_basic_to_ndarray(self):
        frame = VideoFrame(640, 480, "rgb24")
        array = frame.to_ndarray()
        self.assertEqual(array.shape, (480, 640, 3))

    def test_ndarray_gray(self):
        array = numpy.random.randint(0, 256, size=(480, 640), dtype=numpy.uint8)
        for format in ["gray", "gray8"]:
            frame = VideoFrame.from_ndarray(array, format=format)
            self.assertEqual(frame.width, 640)
            self.assertEqual(frame.height, 480)
            self.assertEqual(frame.format.name, "gray")
            self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_gray_align(self):
        array = numpy.random.randint(0, 256, size=(238, 318), dtype=numpy.uint8)
        for format in ["gray", "gray8"]:
            frame = VideoFrame.from_ndarray(array, format=format)
            self.assertEqual(frame.width, 318)
            self.assertEqual(frame.height, 238)
            self.assertEqual(frame.format.name, "gray")
            self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_rgb(self):
        array = numpy.random.randint(0, 256, size=(480, 640, 3), dtype=numpy.uint8)
        for format in ["rgb24", "bgr24"]:
            frame = VideoFrame.from_ndarray(array, format=format)
            self.assertEqual(frame.width, 640)
            self.assertEqual(frame.height, 480)
            self.assertEqual(frame.format.name, format)
            self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_rgb_align(self):
        array = numpy.random.randint(0, 256, size=(238, 318, 3), dtype=numpy.uint8)
        for format in ["rgb24", "bgr24"]:
            frame = VideoFrame.from_ndarray(array, format=format)
            self.assertEqual(frame.width, 318)
            self.assertEqual(frame.height, 238)
            self.assertEqual(frame.format.name, format)
            self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_rgba(self):
        array = numpy.random.randint(0, 256, size=(480, 640, 4), dtype=numpy.uint8)
        for format in ["argb", "rgba", "abgr", "bgra"]:
            frame = VideoFrame.from_ndarray(array, format=format)
            self.assertEqual(frame.width, 640)
            self.assertEqual(frame.height, 480)
            self.assertEqual(frame.format.name, format)
            self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_rgba_align(self):
        array = numpy.random.randint(0, 256, size=(238, 318, 4), dtype=numpy.uint8)
        for format in ["argb", "rgba", "abgr", "bgra"]:
            frame = VideoFrame.from_ndarray(array, format=format)
            self.assertEqual(frame.width, 318)
            self.assertEqual(frame.height, 238)
            self.assertEqual(frame.format.name, format)
            self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_gbrp(self):
        array = numpy.random.randint(0, 256, size=(480, 640, 3), dtype=numpy.uint8)
        frame = VideoFrame.from_ndarray(array, format="gbrp")
        self.assertEqual(frame.width, 640)
        self.assertEqual(frame.height, 480)
        self.assertEqual(frame.format.name, "gbrp")
        self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_gbrp_align(self):
        array = numpy.random.randint(0, 256, size=(238, 318, 3), dtype=numpy.uint8)
        frame = VideoFrame.from_ndarray(array, format="gbrp")
        self.assertEqual(frame.width, 318)
        self.assertEqual(frame.height, 238)
        self.assertEqual(frame.format.name, "gbrp")
        self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_gbrp10(self):
        array = numpy.random.randint(0, 1024, size=(480, 640, 3), dtype=numpy.uint16)
        for format in ["gbrp10be", "gbrp10le"]:
            frame = VideoFrame.from_ndarray(array, format=format)
            self.assertEqual(frame.width, 640)
            self.assertEqual(frame.height, 480)
            self.assertEqual(frame.format.name, format)
            self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_gbrp10_align(self):
        array = numpy.random.randint(0, 1024, size=(238, 318, 3), dtype=numpy.uint16)
        for format in ["gbrp10be", "gbrp10le"]:
            frame = VideoFrame.from_ndarray(array, format=format)
            self.assertEqual(frame.width, 318)
            self.assertEqual(frame.height, 238)
            self.assertEqual(frame.format.name, format)
            self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_gbrp12(self):
        array = numpy.random.randint(0, 4096, size=(480, 640, 3), dtype=numpy.uint16)
        for format in ["gbrp12be", "gbrp12le"]:
            frame = VideoFrame.from_ndarray(array, format=format)
            self.assertEqual(frame.width, 640)
            self.assertEqual(frame.height, 480)
            self.assertEqual(frame.format.name, format)
            self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_gbrp12_align(self):
        array = numpy.random.randint(0, 4096, size=(238, 318, 3), dtype=numpy.uint16)
        for format in ["gbrp12be", "gbrp12le"]:
            frame = VideoFrame.from_ndarray(array, format=format)
            self.assertEqual(frame.width, 318)
            self.assertEqual(frame.height, 238)
            self.assertEqual(frame.format.name, format)
            self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_gbrp14(self):
        array = numpy.random.randint(0, 16384, size=(480, 640, 3), dtype=numpy.uint16)
        for format in ["gbrp14be", "gbrp14le"]:
            frame = VideoFrame.from_ndarray(array, format=format)
            self.assertEqual(frame.width, 640)
            self.assertEqual(frame.height, 480)
            self.assertEqual(frame.format.name, format)
            self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_gbrp14_align(self):
        array = numpy.random.randint(0, 16384, size=(238, 318, 3), dtype=numpy.uint16)
        for format in ["gbrp14be", "gbrp14le"]:
            frame = VideoFrame.from_ndarray(array, format=format)
            self.assertEqual(frame.width, 318)
            self.assertEqual(frame.height, 238)
            self.assertEqual(frame.format.name, format)
            self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_gbrp16(self):
        array = numpy.random.randint(0, 65536, size=(480, 640, 3), dtype=numpy.uint16)
        for format in ["gbrp16be", "gbrp16le"]:
            frame = VideoFrame.from_ndarray(array, format=format)
            self.assertEqual(frame.width, 640)
            self.assertEqual(frame.height, 480)
            self.assertEqual(frame.format.name, format)
            self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_gbrp16_align(self):
        array = numpy.random.randint(0, 65536, size=(238, 318, 3), dtype=numpy.uint16)
        for format in ["gbrp16be", "gbrp16le"]:
            frame = VideoFrame.from_ndarray(array, format=format)
            self.assertEqual(frame.width, 318)
            self.assertEqual(frame.height, 238)
            self.assertEqual(frame.format.name, format)
            self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_gbrpf32(self):
        array = numpy.random.random_sample(size=(480, 640, 3)).astype(numpy.float32)
        for format in ["gbrpf32be", "gbrpf32le"]:
            frame = VideoFrame.from_ndarray(array, format=format)
            self.assertEqual(frame.width, 640)
            self.assertEqual(frame.height, 480)
            self.assertEqual(frame.format.name, format)
            self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_gbrpf32_align(self):
        array = numpy.random.random_sample(size=(238, 318, 3)).astype(numpy.float32)
        for format in ["gbrpf32be", "gbrpf32le"]:
            frame = VideoFrame.from_ndarray(array, format=format)
            self.assertEqual(frame.width, 318)
            self.assertEqual(frame.height, 238)
            self.assertEqual(frame.format.name, format)
            self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_yuv420p(self):
        array = numpy.random.randint(0, 256, size=(720, 640), dtype=numpy.uint8)
        frame = VideoFrame.from_ndarray(array, format="yuv420p")
        self.assertEqual(frame.width, 640)
        self.assertEqual(frame.height, 480)
        self.assertEqual(frame.format.name, "yuv420p")
        self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_yuv420p_align(self):
        array = numpy.random.randint(0, 256, size=(357, 318), dtype=numpy.uint8)
        frame = VideoFrame.from_ndarray(array, format="yuv420p")
        self.assertEqual(frame.width, 318)
        self.assertEqual(frame.height, 238)
        self.assertEqual(frame.format.name, "yuv420p")
        self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_yuvj420p(self):
        array = numpy.random.randint(0, 256, size=(720, 640), dtype=numpy.uint8)
        frame = VideoFrame.from_ndarray(array, format="yuvj420p")
        self.assertEqual(frame.width, 640)
        self.assertEqual(frame.height, 480)
        self.assertEqual(frame.format.name, "yuvj420p")
        self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_yuyv422(self):
        array = numpy.random.randint(0, 256, size=(480, 640, 2), dtype=numpy.uint8)
        frame = VideoFrame.from_ndarray(array, format="yuyv422")
        self.assertEqual(frame.width, 640)
        self.assertEqual(frame.height, 480)
        self.assertEqual(frame.format.name, "yuyv422")
        self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_yuyv422_align(self):
        array = numpy.random.randint(0, 256, size=(238, 318, 2), dtype=numpy.uint8)
        frame = VideoFrame.from_ndarray(array, format="yuyv422")
        self.assertEqual(frame.width, 318)
        self.assertEqual(frame.height, 238)
        self.assertEqual(frame.format.name, "yuyv422")
        self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_gray16be(self):
        array = numpy.random.randint(0, 65536, size=(480, 640), dtype=numpy.uint16)
        frame = VideoFrame.from_ndarray(array, format="gray16be")
        self.assertEqual(frame.width, 640)
        self.assertEqual(frame.height, 480)
        self.assertEqual(frame.format.name, "gray16be")
        self.assertNdarraysEqual(frame.to_ndarray(), array)

        # check endianness by examining value of first pixel
        self.assertPixelValue16(frame.planes[0], array[0][0], "big")

    def test_ndarray_gray16le(self):
        array = numpy.random.randint(0, 65536, size=(480, 640), dtype=numpy.uint16)
        frame = VideoFrame.from_ndarray(array, format="gray16le")
        self.assertEqual(frame.width, 640)
        self.assertEqual(frame.height, 480)
        self.assertEqual(frame.format.name, "gray16le")
        self.assertNdarraysEqual(frame.to_ndarray(), array)

        # check endianness by examining value of first pixel
        self.assertPixelValue16(frame.planes[0], array[0][0], "little")

    def test_ndarray_rgb48be(self):
        array = numpy.random.randint(0, 65536, size=(480, 640, 3), dtype=numpy.uint16)
        frame = VideoFrame.from_ndarray(array, format="rgb48be")
        self.assertEqual(frame.width, 640)
        self.assertEqual(frame.height, 480)
        self.assertEqual(frame.format.name, "rgb48be")
        self.assertNdarraysEqual(frame.to_ndarray(), array)

        # check endianness by examining red value of first pixel
        self.assertPixelValue16(frame.planes[0], array[0][0][0], "big")

    def test_ndarray_rgb48le(self):
        array = numpy.random.randint(0, 65536, size=(480, 640, 3), dtype=numpy.uint16)
        frame = VideoFrame.from_ndarray(array, format="rgb48le")
        self.assertEqual(frame.width, 640)
        self.assertEqual(frame.height, 480)
        self.assertEqual(frame.format.name, "rgb48le")
        self.assertNdarraysEqual(frame.to_ndarray(), array)

        # check endianness by examining red value of first pixel
        self.assertPixelValue16(frame.planes[0], array[0][0][0], "little")

    def test_ndarray_rgb48le_align(self):
        array = numpy.random.randint(0, 65536, size=(238, 318, 3), dtype=numpy.uint16)
        frame = VideoFrame.from_ndarray(array, format="rgb48le")
        self.assertEqual(frame.width, 318)
        self.assertEqual(frame.height, 238)
        self.assertEqual(frame.format.name, "rgb48le")
        self.assertNdarraysEqual(frame.to_ndarray(), array)

        # check endianness by examining red value of first pixel
        self.assertPixelValue16(frame.planes[0], array[0][0][0], "little")

    def test_ndarray_rgba64be(self):
        array = numpy.random.randint(0, 65536, size=(480, 640, 4), dtype=numpy.uint16)
        frame = VideoFrame.from_ndarray(array, format="rgba64be")
        self.assertEqual(frame.width, 640)
        self.assertEqual(frame.height, 480)
        self.assertEqual(frame.format.name, "rgba64be")
        self.assertNdarraysEqual(frame.to_ndarray(), array)

        # check endianness by examining red value of first pixel
        self.assertPixelValue16(frame.planes[0], array[0][0][0], "big")

    def test_ndarray_rgba64le(self):
        array = numpy.random.randint(0, 65536, size=(480, 640, 4), dtype=numpy.uint16)
        frame = VideoFrame.from_ndarray(array, format="rgba64le")
        self.assertEqual(frame.width, 640)
        self.assertEqual(frame.height, 480)
        self.assertEqual(frame.format.name, "rgba64le")
        self.assertNdarraysEqual(frame.to_ndarray(), array)

        # check endianness by examining red value of first pixel
        self.assertPixelValue16(frame.planes[0], array[0][0][0], "little")

    def test_ndarray_rgb8(self):
        array = numpy.random.randint(0, 256, size=(480, 640), dtype=numpy.uint8)
        frame = VideoFrame.from_ndarray(array, format="rgb8")
        self.assertEqual(frame.width, 640)
        self.assertEqual(frame.height, 480)
        self.assertEqual(frame.format.name, "rgb8")
        self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_bgr8(self):
        array = numpy.random.randint(0, 256, size=(480, 640), dtype=numpy.uint8)
        frame = VideoFrame.from_ndarray(array, format="bgr8")
        self.assertEqual(frame.width, 640)
        self.assertEqual(frame.height, 480)
        self.assertEqual(frame.format.name, "bgr8")
        self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_pal8(self):
        array = numpy.random.randint(0, 256, size=(480, 640), dtype=numpy.uint8)
        palette = numpy.random.randint(0, 256, size=(256, 4), dtype=numpy.uint8)
        frame = VideoFrame.from_ndarray((array, palette), format="pal8")
        self.assertEqual(frame.width, 640)
        self.assertEqual(frame.height, 480)
        self.assertEqual(frame.format.name, "pal8")
        returned = frame.to_ndarray()
        self.assertTrue((type(returned) is tuple) and len(returned) == 2)
        self.assertNdarraysEqual(returned[0], array)
        self.assertNdarraysEqual(returned[1], palette)

    def test_ndarray_nv12(self):
        array = numpy.random.randint(0, 256, size=(720, 640), dtype=numpy.uint8)
        frame = VideoFrame.from_ndarray(array, format="nv12")
        self.assertEqual(frame.width, 640)
        self.assertEqual(frame.height, 480)
        self.assertEqual(frame.format.name, "nv12")
        self.assertNdarraysEqual(frame.to_ndarray(), array)

    def test_ndarray_nv12_align(self):
        array = numpy.random.randint(0, 256, size=(357, 318), dtype=numpy.uint8)
        frame = VideoFrame.from_ndarray(array, format="nv12")
        self.assertEqual(frame.width, 318)
        self.assertEqual(frame.height, 238)
        self.assertEqual(frame.format.name, "nv12")
        self.assertNdarraysEqual(frame.to_ndarray(), array)


class TestVideoFrameTiming(TestCase):
    def test_reformat_pts(self):
        frame = VideoFrame(640, 480, "rgb24")
        frame.pts = 123
        frame.time_base = "456/1"  # Just to be different.
        frame = frame.reformat(320, 240)
        self.assertEqual(frame.pts, 123)
        self.assertEqual(frame.time_base, 456)


class TestVideoFrameReformat(TestCase):
    def test_reformat_identity(self):
        frame1 = VideoFrame(640, 480, "rgb24")
        frame2 = frame1.reformat(640, 480, "rgb24")
        self.assertIs(frame1, frame2)

    def test_reformat_colourspace(self):

        # This is allowed.
        frame = VideoFrame(640, 480, "rgb24")
        frame.reformat(src_colorspace=None, dst_colorspace="smpte240")

        # I thought this was not allowed, but it seems to be.
        frame = VideoFrame(640, 480, "yuv420p")
        frame.reformat(src_colorspace=None, dst_colorspace="smpte240")

    def test_reformat_pixel_format_align(self):
        height = 480
        for width in range(2, 258, 2):
            frame_yuv = VideoFrame(width, height, "yuv420p")
            for plane in frame_yuv.planes:
                plane.update(b"\xff" * plane.buffer_size)

            expected_rgb = numpy.zeros(shape=(height, width, 3), dtype=numpy.uint8)
            expected_rgb[:, :, 0] = 255
            expected_rgb[:, :, 1] = 124
            expected_rgb[:, :, 2] = 255

            frame_rgb = frame_yuv.reformat(format="rgb24")
            self.assertNdarraysEqual(frame_rgb.to_ndarray(), expected_rgb)



from fractions import Fraction
import warnings

from av.audio.format cimport AudioFormat
from av.audio.frame cimport AudioFrame
from av.audio.layout cimport AudioLayout
from av.error cimport err_check
from av.filter.context cimport FilterContext, wrap_filter_context
from av.filter.filter cimport Filter, wrap_filter
from av.video.format cimport VideoFormat
from av.video.frame cimport VideoFrame


cdef class Graph(object):

    def __cinit__(self):

        self.ptr = lib.avfilter_graph_alloc()
        self.configured = False
        self._name_counts = {}

        self._nb_filters_seen = 0
        self._context_by_ptr = {}
        self._context_by_name = {}
        self._context_by_type = {}

    def __dealloc__(self):
        if self.ptr:
            # This frees the graph, filter contexts, links, etc..
            lib.avfilter_graph_free(&self.ptr)

    cdef str _get_unique_name(self, str name):
        count = self._name_counts.get(name, 0)
        self._name_counts[name] = count + 1
        if count:
            return '%s_%s' % (name, count)
        else:
            return name

    cpdef configure(self, bint auto_buffer=True, bint force=False):
        if self.configured and not force:
            return

        # if auto_buffer:
        #     for ctx in self._context_by_ptr.itervalues():
        #         for in_ in ctx.inputs:
        #             if not in_.link:
        #                 if in_.type == 'video':
        #                     pass

        err_check(lib.avfilter_graph_config(self.ptr, NULL))
        self.configured = True

        # We get auto-inserted stuff here.
        self._auto_register()

    # def parse_string(self, str filter_str):
        # err_check(lib.avfilter_graph_parse2(self.ptr, filter_str, &self.inputs, &self.outputs))
        #
        # cdef lib.AVFilterInOut *input_
        # while input_ != NULL:
        #     print 'in ', input_.pad_idx, (input_.name if input_.name != NULL else ''), input_.filter_ctx.name, input_.filter_ctx.filter.name
        #     input_ = input_.next
        #
        # cdef lib.AVFilterInOut *output
        # while output != NULL:
        #     print 'out', output.pad_idx, (output.name if output.name != NULL else ''), output.filter_ctx.name, output.filter_ctx.filter.name
        #     output = output.next

    # NOTE: Only FFmpeg supports this.
    # def dump(self):
    #     cdef char *buf = lib.avfilter_graph_dump(self.ptr, "")
    #     cdef str ret = buf
    #     lib.av_free(buf)
    #     return ret

    def add(self, filter, args=None, **kwargs):

        cdef Filter cy_filter
        if isinstance(filter, str):
            cy_filter = Filter(filter)
        elif isinstance(filter, Filter):
            cy_filter = filter
        else:
            raise TypeError("filter must be a string or Filter")

        cdef str name = self._get_unique_name(kwargs.pop('name', None) or cy_filter.name)

        cdef lib.AVFilterContext *ptr = lib.avfilter_graph_alloc_filter(self.ptr, cy_filter.ptr, name)
        if not ptr:
            raise RuntimeError("Could not allocate AVFilterContext")

        # Manually construct this context (so we can return it).
        cdef FilterContext ctx = wrap_filter_context(self, cy_filter, ptr)
        ctx.init(args, **kwargs)
        self._register_context(ctx)

        # There might have been automatic contexts added (e.g. resamplers,
        # fifos, and scalers). It is more likely to see them after the graph
        # is configured, but we wan't to be safe.
        self._auto_register()

        return ctx

    cdef _register_context(self, FilterContext ctx):
        self._context_by_ptr[<long>ctx.ptr] = ctx
        self._context_by_name[ctx.ptr.name] = ctx
        self._context_by_type.setdefault(ctx.filter.ptr.name, []).append(ctx)

    cdef _auto_register(self):
        cdef int i
        cdef lib.AVFilterContext *c_ctx
        cdef Filter filter_
        cdef FilterContext py_ctx
        # We assume that filters are never removed from the graph. At this
        # point we don't expose that in the API, so we should be okay...
        for i in range(self._nb_filters_seen, self.ptr.nb_filters):
            c_ctx = self.ptr.filters[i]
            if <long>c_ctx in self._context_by_ptr:
                continue
            filter_ = wrap_filter(c_ctx.filter)
            py_ctx = wrap_filter_context(self, filter_, c_ctx)
            self._register_context(py_ctx)
        self._nb_filters_seen = self.ptr.nb_filters

    def add_buffer(self, template=None, width=None, height=None, format=None, name=None, time_base=None):

        if template is not None:
            if width is None:
                width = template.width
            if height is None:
                height = template.height
            if format is None:
                format = template.format
            if time_base is None:
                time_base = template.time_base

        if width is None:
            raise ValueError('missing width')
        if height is None:
            raise ValueError('missing height')
        if format is None:
            raise ValueError('missing format')
        if time_base is None:
            warnings.warn('missing time_base. Guessing 1/1000 time base. '
                          'This is deprecated and may be removed in future releases.',
                          DeprecationWarning)
            time_base = Fraction(1, 1000)

        return self.add(
            'buffer',
            name=name,
            video_size=f'{width}x{height}',
            pix_fmt=str(int(VideoFormat(format))),
            time_base=str(time_base),
            pixel_aspect='1/1',
        )

    def add_abuffer(self, template=None, sample_rate=None, format=None, layout=None, channels=None, name=None, time_base=None):
        """
        Convenience method for adding `abuffer <https://ffmpeg.org/ffmpeg-filters.html#abuffer>`_.
        """

        if template is not None:
            if sample_rate is None:
                sample_rate = template.sample_rate
            if format is None:
                format = template.format
            if layout is None:
                layout = template.layout.name
            if channels is None:
                channels = template.channels
            if time_base is None:
                time_base = template.time_base

        if sample_rate is None:
            raise ValueError('missing sample_rate')
        if format is None:
            raise ValueError('missing format')
        if layout is None and channels is None:
            raise ValueError('missing layout or channels')
        if time_base is None:
            time_base = Fraction(1, sample_rate)

        kwargs = dict(
            sample_rate=str(sample_rate),
            sample_fmt=AudioFormat(format).name,
            time_base=str(time_base),
        )
        if layout:
            kwargs['channel_layout'] = AudioLayout(layout).name
        if channels:
            kwargs['channels'] = str(channels)

        return self.add('abuffer', name=name, **kwargs)

    def push(self, frame):

        if frame is None:
            contexts = self._context_by_type.get('buffer', []) + self._context_by_type.get('abuffer', [])
        elif isinstance(frame, VideoFrame):
            contexts = self._context_by_type.get('buffer', [])
        elif isinstance(frame, AudioFrame):
            contexts = self._context_by_type.get('abuffer', [])
        else:
            raise ValueError('can only AudioFrame, VideoFrame or None; got %s' % type(frame))

        if len(contexts) != 1:
            raise ValueError('can only auto-push with single buffer; found %s' % len(contexts))

        contexts[0].push(frame)

    def pull(self):

        vsinks = self._context_by_type.get('buffersink', [])
        asinks = self._context_by_type.get('abuffersink', [])

        nsinks = len(vsinks) + len(asinks)
        if nsinks != 1:
            raise ValueError('can only auto-pull with single sink; found %s' % nsinks)

        return (vsinks or asinks)[0].pull()



from libc.stdint cimport int64_t, uint64_t


cdef extern from "libavformat/avformat.h" nogil:

    cdef int   avformat_version()
    cdef char* avformat_configuration()
    cdef char* avformat_license()
    cdef void  avformat_network_init()

    cdef int64_t INT64_MIN

    cdef int AV_TIME_BASE
    cdef int AVSEEK_FLAG_BACKWARD
    cdef int AVSEEK_FLAG_BYTE
    cdef int AVSEEK_FLAG_ANY
    cdef int AVSEEK_FLAG_FRAME


    cdef int AVIO_FLAG_WRITE

    cdef enum AVMediaType:
        AVMEDIA_TYPE_UNKNOWN
        AVMEDIA_TYPE_VIDEO
        AVMEDIA_TYPE_AUDIO
        AVMEDIA_TYPE_DATA
        AVMEDIA_TYPE_SUBTITLE
        AVMEDIA_TYPE_ATTACHMENT
        AVMEDIA_TYPE_NB

    cdef struct AVStream:

        int index
        int id

        AVCodecParameters *codecpar

        AVRational time_base

        int64_t start_time
        int64_t duration
        int64_t nb_frames
        int64_t cur_dts

        AVDictionary *metadata

        AVRational avg_frame_rate
        AVRational r_frame_rate
        AVRational sample_aspect_ratio

    # http://ffmpeg.org/doxygen/trunk/structAVIOContext.html
    cdef struct AVIOContext:
        unsigned char* buffer
        int buffer_size
        int write_flag
        int direct
        int seekable
        int max_packet_size
        void *opaque

    # http://ffmpeg.org/doxygen/trunk/structAVIOInterruptCB.html
    cdef struct AVIOInterruptCB:
        int (*callback)(void*)
        void *opaque

    cdef int AVIO_FLAG_DIRECT
    cdef int AVIO_SEEKABLE_NORMAL

    cdef int SEEK_SET
    cdef int SEEK_CUR
    cdef int SEEK_END
    cdef int AVSEEK_SIZE

    cdef AVIOContext* avio_alloc_context(
        unsigned char *buffer,
        int buffer_size,
        int write_flag,
        void *opaque,
        int(*read_packet)(void *opaque, uint8_t *buf, int buf_size),
        int(*write_packet)(void *opaque, uint8_t *buf, int buf_size),
        int64_t(*seek)(void *opaque, int64_t offset, int whence)
    )

    # http://ffmpeg.org/doxygen/trunk/structAVInputFormat.html
    cdef struct AVInputFormat:
        const char *name
        const char *long_name
        const char *extensions
        int flags
        # const AVCodecTag* const *codec_tag
        const AVClass *priv_class

    cdef struct AVProbeData:
        unsigned char *buf
        int buf_size
        const char *filename

    cdef AVInputFormat* av_probe_input_format(
        AVProbeData *pd,
        int is_opened
    )

    # http://ffmpeg.org/doxygen/trunk/structAVOutputFormat.html
    cdef struct AVOutputFormat:
        const char *name
        const char *long_name
        const char *extensions
        AVCodecID video_codec
        AVCodecID audio_codec
        AVCodecID subtitle_codec
        int flags
        # const AVCodecTag* const *codec_tag
        const AVClass *priv_class

    # AVInputFormat.flags and AVOutputFormat.flags
    cdef enum:
        AVFMT_NOFILE
        AVFMT_NEEDNUMBER
        AVFMT_SHOW_IDS
        AVFMT_GLOBALHEADER
        AVFMT_NOTIMESTAMPS
        AVFMT_GENERIC_INDEX
        AVFMT_TS_DISCONT
        AVFMT_VARIABLE_FPS
        AVFMT_NODIMENSIONS
        AVFMT_NOSTREAMS
        AVFMT_NOBINSEARCH
        AVFMT_NOGENSEARCH
        AVFMT_NO_BYTE_SEEK
        AVFMT_ALLOW_FLUSH
        AVFMT_TS_NONSTRICT
        AVFMT_TS_NEGATIVE
        AVFMT_SEEK_TO_PTS

    # AVFormatContext.flags
    cdef enum:
        AVFMT_FLAG_GENPTS
        AVFMT_FLAG_IGNIDX
        AVFMT_FLAG_NONBLOCK
        AVFMT_FLAG_IGNDTS
        AVFMT_FLAG_NOFILLIN
        AVFMT_FLAG_NOPARSE
        AVFMT_FLAG_NOBUFFER
        AVFMT_FLAG_CUSTOM_IO
        AVFMT_FLAG_DISCARD_CORRUPT
        AVFMT_FLAG_FLUSH_PACKETS
        AVFMT_FLAG_BITEXACT
        AVFMT_FLAG_SORT_DTS
        AVFMT_FLAG_FAST_SEEK
        AVFMT_FLAG_SHORTEST
        AVFMT_FLAG_AUTO_BSF

    cdef int av_probe_input_buffer(
        AVIOContext *pb,
        AVInputFormat **fmt,
        const char *filename,
        void *logctx,
        unsigned int offset,
        unsigned int max_probe_size
    )

    cdef AVInputFormat* av_find_input_format(const char *name)

    # http://ffmpeg.org/doxygen/trunk/structAVFormatContext.html
    cdef struct AVFormatContext:

        # Streams.
        unsigned int nb_streams
        AVStream **streams

        AVInputFormat *iformat
        AVOutputFormat *oformat

        AVIOContext *pb
        AVIOInterruptCB interrupt_callback

        AVDictionary *metadata

        char filename
        int64_t start_time
        int64_t duration
        int bit_rate

        int flags
        int64_t max_analyze_duration

        void *opaque

        int (*io_open)(
            AVFormatContext *s,
            AVIOContext **pb,
            const char *url,
            int flags,
            AVDictionary **options
        )
        void (*io_close)(
            AVFormatContext *s,
            AVIOContext *pb
        )

    cdef AVFormatContext* avformat_alloc_context()

    # .. c:function:: avformat_open_input(...)
    #
    #       Options are passed via :func:`av.open`.
    #
    #       .. seealso:: FFmpeg's docs: :ffmpeg:`avformat_open_input`
    #
    cdef int avformat_open_input(
        AVFormatContext **ctx, # NULL will allocate for you.
        char *filename,
        AVInputFormat *format, # Can be NULL.
        AVDictionary **options # Can be NULL.
    )

    cdef int avformat_close_input(AVFormatContext **ctx)

    # .. c:function:: avformat_write_header(...)
    #
    #       Options are passed via :func:`av.open`; called in
    #       :meth:`av.container.OutputContainer.start_encoding`.
    #
    #       .. seealso:: FFmpeg's docs: :ffmpeg:`avformat_write_header`
    #
    cdef int avformat_write_header(
        AVFormatContext *ctx,
        AVDictionary **options # Can be NULL
    )

    cdef int av_write_trailer(AVFormatContext *ctx)

    cdef int av_interleaved_write_frame(
        AVFormatContext *ctx,
        AVPacket *pkt
    )

    cdef int av_write_frame(
        AVFormatContext *ctx,
        AVPacket *pkt
    )

    cdef int avio_open(
        AVIOContext **s,
        char *url,
        int flags
    )

    cdef int64_t avio_size(
        AVIOContext *s
    )

    cdef AVOutputFormat* av_guess_format(
        char *short_name,
        char *filename,
        char *mime_type
    )

    cdef int avformat_query_codec(
        AVOutputFormat *ofmt,
        AVCodecID codec_id,
        int std_compliance
    )

    cdef void avio_flush(AVIOContext *s)

    cdef int avio_close(AVIOContext *s)

    cdef int avio_closep(AVIOContext **s)

    cdef int avformat_find_stream_info(
        AVFormatContext *ctx,
        AVDictionary **options, # Can be NULL.
    )

    cdef AVStream* avformat_new_stream(
        AVFormatContext *ctx,
        AVCodec *c
    )

    cdef int avformat_alloc_output_context2(
        AVFormatContext **ctx,
        AVOutputFormat *oformat,
        char *format_name,
        char *filename
    )

    cdef int avformat_free_context(AVFormatContext *ctx)

    cdef AVClass* avformat_get_class()

    cdef void av_dump_format(
        AVFormatContext *ctx,
        int index,
        char *url,
        int is_output,
    )

    cdef int av_read_frame(
        AVFormatContext *ctx,
        AVPacket *packet,
    )

    cdef int av_seek_frame(
        AVFormatContext *ctx,
        int stream_index,
        int64_t timestamp,
        int flags
    )

    cdef int avformat_seek_file(
        AVFormatContext *ctx,
        int stream_index,
        int64_t min_ts,
        int64_t ts,
        int64_t max_ts,
        int flags
    )

    cdef AVRational av_guess_frame_rate(
        AVFormatContext *ctx,
        AVStream *stream,
        AVFrame *frame
    )

    cdef const AVInputFormat* av_demuxer_iterate(void **opaque)
    cdef const AVOutputFormat* av_muxer_iterate(void **opaque)

    # custom

    cdef set pyav_get_available_formats()



cimport libav as lib

from av.error cimport err_check
from av.frame cimport Frame
from av.packet cimport Packet
from av.subtitles.subtitle cimport SubtitleProxy, SubtitleSet


cdef class SubtitleCodecContext(CodecContext):

    cdef _send_packet_and_recv(self, Packet packet):
        cdef SubtitleProxy proxy = SubtitleProxy()

        cdef int got_frame = 0
        err_check(lib.avcodec_decode_subtitle2(
            self.ptr,
            &proxy.struct,
            &got_frame,
            packet.ptr if packet else NULL))
        if got_frame:
            return [SubtitleSet(proxy)]
        else:
            return []



cimport libav as lib

from av.filter.graph cimport Graph
from av.filter.pad cimport FilterContextPad


cdef class FilterLink(object):

    cdef readonly Graph graph
    cdef lib.AVFilterLink *ptr

    cdef FilterContextPad _input
    cdef FilterContextPad _output


cdef FilterLink wrap_filter_link(Graph graph, lib.AVFilterLink *ptr)



from cpython.buffer cimport Py_buffer


cdef class ByteSource(object):

    cdef object owner

    cdef bint has_view
    cdef Py_buffer view

    cdef unsigned char *ptr
    cdef size_t length

cdef ByteSource bytesource(object, bint allow_none=*)



from libc.stdint cimport uint8_t
cimport libav as lib

from av.enum cimport define_enum
from av.error cimport err_check
from av.video.format cimport VideoFormat
from av.video.frame cimport alloc_video_frame


Interpolation = define_enum('Interpolation', __name__, (
    ('FAST_BILINEAR', lib.SWS_FAST_BILINEAR, "Fast bilinear"),
    ('BILINEAR', lib.SWS_BILINEAR, "Bilinear"),
    ('BICUBIC', lib.SWS_BICUBIC, "Bicubic"),
    ('X', lib.SWS_X, "Experimental"),
    ('POINT', lib.SWS_POINT, "Nearest neighbor / point"),
    ('AREA', lib.SWS_AREA, "Area averaging"),
    ('BICUBLIN', lib.SWS_BICUBLIN, "Luma bicubic / chroma bilinear"),
    ('GAUSS', lib.SWS_GAUSS, "Gaussian"),
    ('SINC', lib.SWS_SINC, "Sinc"),
    ('LANCZOS', lib.SWS_LANCZOS, "Lanczos"),
    ('SPLINE', lib.SWS_SPLINE, "Bicubic spline"),
))

Colorspace = define_enum('Colorspace', __name__, (

    ('ITU709', lib.SWS_CS_ITU709),
    ('FCC', lib.SWS_CS_FCC),
    ('ITU601', lib.SWS_CS_ITU601),
    ('ITU624', lib.SWS_CS_ITU624),
    ('SMPTE170M', lib.SWS_CS_SMPTE170M),
    ('SMPTE240M', lib.SWS_CS_SMPTE240M),
    ('DEFAULT', lib.SWS_CS_DEFAULT),

    # Lowercase for b/c.
    ('itu709', lib.SWS_CS_ITU709),
    ('fcc', lib.SWS_CS_FCC),
    ('itu601', lib.SWS_CS_ITU601),
    ('itu624', lib.SWS_CS_SMPTE170M),
    ('smpte240', lib.SWS_CS_SMPTE240M),
    ('default', lib.SWS_CS_DEFAULT),

))


cdef class VideoReformatter(object):

    """An object for reformatting size and pixel format of :class:`.VideoFrame`.

    It is most efficient to have a reformatter object for each set of parameters
    you will use as calling :meth:`reformat` will reconfigure the internal object.

    """

    def __dealloc__(self):
        with nogil:
            lib.sws_freeContext(self.ptr)

    def reformat(self, VideoFrame frame not None, width=None, height=None,
                 format=None, src_colorspace=None, dst_colorspace=None,
                 interpolation=None):
        """Create a new :class:`VideoFrame` with the given width/height/format/colorspace.

        Returns the same frame untouched if nothing needs to be done to it.

        :param int width: New width, or ``None`` for the same width.
        :param int height: New height, or ``None`` for the same height.
        :param format: New format, or ``None`` for the same format.
        :type  format: :class:`.VideoFormat` or ``str``
        :param src_colorspace: Current colorspace, or ``None`` for ``DEFAULT``.
        :type  src_colorspace: :class:`Colorspace` or ``str``
        :param dst_colorspace: Desired colorspace, or ``None`` for ``DEFAULT``.
        :type  dst_colorspace: :class:`Colorspace` or ``str``
        :param interpolation: The interpolation method to use, or ``None`` for ``BILINEAR``.
        :type  interpolation: :class:`Interpolation` or ``str``

        """

        cdef VideoFormat video_format = VideoFormat(format if format is not None else frame.format)
        cdef int c_src_colorspace = (Colorspace[src_colorspace] if src_colorspace is not None else Colorspace.DEFAULT).value
        cdef int c_dst_colorspace = (Colorspace[dst_colorspace] if dst_colorspace is not None else Colorspace.DEFAULT).value
        cdef int c_interpolation = (Interpolation[interpolation] if interpolation is not None else Interpolation.BILINEAR).value

        return self._reformat(
            frame,
            width or frame.ptr.width,
            height or frame.ptr.height,
            video_format.pix_fmt,
            c_src_colorspace,
            c_dst_colorspace,
            c_interpolation,
        )

    cdef _reformat(self, VideoFrame frame, int width, int height,
                   lib.AVPixelFormat dst_format, int src_colorspace,
                   int dst_colorspace, int interpolation):

        if frame.ptr.format < 0:
            raise ValueError("Frame does not have format set.")

        cdef lib.AVPixelFormat src_format = <lib.AVPixelFormat> frame.ptr.format

        # Shortcut!
        if (
            dst_format == src_format and
            width == frame.ptr.width and
            height == frame.ptr.height and
            dst_colorspace == src_colorspace
        ):
            return frame

        # Try and reuse existing SwsContextProxy
        # VideoStream.decode will copy its SwsContextProxy to VideoFrame
        # So all Video frames from the same VideoStream should have the same one
        with nogil:
            self.ptr = lib.sws_getCachedContext(
                self.ptr,
                frame.ptr.width,
                frame.ptr.height,
                src_format,
                width,
                height,
                dst_format,
                interpolation,
                NULL,
                NULL,
                NULL
            )

        # We want to change the colorspace transforms. We do that by grabbing
        # all of the current settings, changing a couple, and setting them all.
        # We need a lot of state here.
        cdef const int *inv_tbl
        cdef const int *tbl
        cdef int src_range, dst_range, brightness, contrast, saturation
        cdef int ret
        if src_colorspace != dst_colorspace:

            with nogil:

                # Casts for const-ness, because Cython isn't expressive enough.
                ret = lib.sws_getColorspaceDetails(
                    self.ptr,
                    <int**>&inv_tbl,
                    &src_range,
                    <int**>&tbl,
                    &dst_range,
                    &brightness,
                    &contrast,
                    &saturation
                )

            err_check(ret)

            with nogil:

                # Grab the coefficients for the requested transforms.
                # The inv_table brings us to linear, and `tbl` to the new space.
                if src_colorspace != lib.SWS_CS_DEFAULT:
                    inv_tbl = lib.sws_getCoefficients(src_colorspace)
                if dst_colorspace != lib.SWS_CS_DEFAULT:
                    tbl = lib.sws_getCoefficients(dst_colorspace)

                # Apply!
                ret = lib.sws_setColorspaceDetails(
                    self.ptr,
                    inv_tbl,
                    src_range,
                    tbl,
                    dst_range,
                    brightness,
                    contrast,
                    saturation
                )

            err_check(ret)

        # Create a new VideoFrame.
        cdef VideoFrame new_frame = alloc_video_frame()
        new_frame._copy_internal_attributes(frame)
        new_frame._init(dst_format, width, height)

        # Finally, scale the image.
        with nogil:
            lib.sws_scale(
                self.ptr,
                # Cast for const-ness, because Cython isn't expressive enough.
                <const uint8_t**>frame.ptr.data,
                frame.ptr.linesize,
                0,  # slice Y
                frame.ptr.height,
                new_frame.ptr.data,
                new_frame.ptr.linesize,
            )

        return new_frame



from fractions import Fraction
from unittest import SkipTest
import math

from av import AudioFrame, VideoFrame
from av.audio.stream import AudioStream
from av.video.stream import VideoStream
import av

from .common import Image, TestCase, fate_suite


WIDTH = 320
HEIGHT = 240
DURATION = 48


def write_rgb_rotate(output):

    if not Image:
        raise SkipTest()

    output.metadata["title"] = "container"
    output.metadata["key"] = "value"

    stream = output.add_stream("mpeg4", 24)
    stream.width = WIDTH
    stream.height = HEIGHT
    stream.pix_fmt = "yuv420p"

    for frame_i in range(DURATION):

        frame = VideoFrame(WIDTH, HEIGHT, "rgb24")
        image = Image.new(
            "RGB",
            (WIDTH, HEIGHT),
            (
                int(255 * (0.5 + 0.5 * math.sin(frame_i / DURATION * 2 * math.pi))),
                int(
                    255
                    * (
                        0.5
                        + 0.5
                        * math.sin(frame_i / DURATION * 2 * math.pi + 2 / 3 * math.pi)
                    )
                ),
                int(
                    255
                    * (
                        0.5
                        + 0.5
                        * math.sin(frame_i / DURATION * 2 * math.pi + 4 / 3 * math.pi)
                    )
                ),
            ),
        )
        frame.planes[0].update(image.tobytes())

        for packet in stream.encode(frame):
            output.mux(packet)

    for packet in stream.encode(None):
        output.mux(packet)


def assert_rgb_rotate(self, input_, is_dash=False):

    # Now inspect it a little.
    self.assertEqual(len(input_.streams), 1)
    if is_dash:
        # FFmpeg 4.2 added parsing of the programme information and it is named "Title"
        if av.library_versions["libavformat"] >= (58, 28):
            self.assertTrue(
                input_.metadata.get("Title") == "container", input_.metadata
            )
    else:
        self.assertEqual(input_.metadata.get("title"), "container", input_.metadata)
    self.assertEqual(input_.metadata.get("key"), None)

    stream = input_.streams[0]

    if is_dash:
        # The DASH format doesn't provide a duration for the stream
        # and so the container duration (micro seconds) is checked instead
        self.assertEqual(input_.duration, 2000000)
        expected_average_rate = 24
        expected_duration = None
        expected_frames = 0
        expected_id = 0
    else:
        if av.library_versions["libavformat"] < (58, 76):
            # FFmpeg < 4.4
            expected_average_rate = Fraction(1152, 47)
            expected_duration = 24064
        else:
            # FFmpeg >= 4.4
            expected_average_rate = 24
            expected_duration = 24576
        expected_frames = 48
        expected_id = 1

    # actual stream properties
    self.assertIsInstance(stream, VideoStream)
    self.assertEqual(stream.average_rate, expected_average_rate)
    self.assertEqual(stream.base_rate, 24)
    self.assertEqual(stream.duration, expected_duration)
    self.assertEqual(stream.guessed_rate, 24)
    self.assertEqual(stream.frames, expected_frames)
    self.assertEqual(stream.id, expected_id)
    self.assertEqual(stream.index, 0)
    self.assertEqual(stream.profile, "Simple Profile")
    self.assertEqual(stream.start_time, 0)
    self.assertEqual(stream.time_base, Fraction(1, 12288))
    self.assertEqual(stream.type, "video")

    # codec context properties
    self.assertEqual(stream.codec.name, "mpeg4")
    self.assertEqual(stream.codec.long_name, "MPEG-4 part 2")
    self.assertEqual(stream.format.name, "yuv420p")
    self.assertEqual(stream.format.width, WIDTH)
    self.assertEqual(stream.format.height, HEIGHT)
    self.assertEqual(stream.ticks_per_frame, 1)


class TestBasicVideoEncoding(TestCase):
    def test_default_options(self):
        with av.open(self.sandboxed("output.mov"), "w") as output:
            stream = output.add_stream("mpeg4")
            self.assertIn(stream, output.streams.video)
            self.assertEqual(stream.average_rate, Fraction(24, 1))
            self.assertEqual(stream.time_base, None)

            # codec context properties
            self.assertEqual(stream.bit_rate, 1024000)
            self.assertEqual(stream.format.height, 480)
            self.assertEqual(stream.format.name, "yuv420p")
            self.assertEqual(stream.format.width, 640)
            self.assertEqual(stream.height, 480)
            self.assertEqual(stream.pix_fmt, "yuv420p")
            self.assertEqual(stream.ticks_per_frame, 1)
            self.assertEqual(stream.width, 640)

    def test_encoding(self):
        path = self.sandboxed("rgb_rotate.mov")

        with av.open(path, "w") as output:
            write_rgb_rotate(output)
        with av.open(path) as input:
            assert_rgb_rotate(self, input)

    def test_encoding_with_pts(self):
        path = self.sandboxed("video_with_pts.mov")

        with av.open(path, "w") as output:
            stream = output.add_stream("libx264", 24)
            self.assertIn(stream, output.streams.video)
            stream.width = WIDTH
            stream.height = HEIGHT
            stream.pix_fmt = "yuv420p"

            for i in range(DURATION):
                frame = VideoFrame(WIDTH, HEIGHT, "rgb24")
                frame.pts = i * 2000
                frame.time_base = Fraction(1, 48000)

                for packet in stream.encode(frame):
                    self.assertEqual(packet.time_base, Fraction(1, 24))
                    output.mux(packet)

            for packet in stream.encode(None):
                self.assertEqual(packet.time_base, Fraction(1, 24))
                output.mux(packet)

    def test_encoding_with_unicode_filename(self):
        path = self.sandboxed("¢∞§¶•ªº.mov")

        with av.open(path, "w") as output:
            write_rgb_rotate(output)
        with av.open(path) as input:
            assert_rgb_rotate(self, input)


class TestBasicAudioEncoding(TestCase):
    def test_default_options(self):
        with av.open(self.sandboxed("output.mov"), "w") as output:
            stream = output.add_stream("mp2")
            self.assertIn(stream, output.streams.audio)
            self.assertEqual(stream.time_base, None)

            # codec context properties
            self.assertEqual(stream.bit_rate, 128000)
            self.assertEqual(stream.format.name, "s16")
            self.assertEqual(stream.sample_rate, 48000)
            self.assertEqual(stream.ticks_per_frame, 1)

    def test_transcode(self):
        path = self.sandboxed("audio_transcode.mov")

        with av.open(path, "w") as output:
            output.metadata["title"] = "container"
            output.metadata["key"] = "value"

            sample_rate = 48000
            channel_layout = "stereo"
            channels = 2
            sample_fmt = "s16"

            stream = output.add_stream("mp2", sample_rate)
            self.assertIn(stream, output.streams.audio)

            ctx = stream.codec_context
            ctx.time_base = sample_rate
            ctx.sample_rate = sample_rate
            ctx.format = sample_fmt
            ctx.layout = channel_layout
            ctx.channels = channels

            with av.open(
                fate_suite("audio-reference/chorusnoise_2ch_44kHz_s16.wav")
            ) as src:
                for frame in src.decode(audio=0):
                    for packet in stream.encode(frame):
                        output.mux(packet)

            for packet in stream.encode(None):
                output.mux(packet)

        with av.open(path) as container:
            self.assertEqual(len(container.streams), 1)
            self.assertEqual(
                container.metadata.get("title"), "container", container.metadata
            )
            self.assertEqual(container.metadata.get("key"), None)

            stream = container.streams[0]
            self.assertIsInstance(stream, AudioStream)

            # codec context properties
            self.assertEqual(stream.channels, channels)
            self.assertEqual(stream.format.name, "s16p")
            self.assertEqual(stream.sample_rate, sample_rate)


class TestEncodeStreamSemantics(TestCase):
    def test_stream_index(self):
        with av.open(self.sandboxed("output.mov"), "w") as output:
            vstream = output.add_stream("mpeg4", 24)
            self.assertIn(vstream, output.streams.video)
            vstream.pix_fmt = "yuv420p"
            vstream.width = 320
            vstream.height = 240

            astream = output.add_stream("mp2", 48000)
            self.assertIn(astream, output.streams.audio)
            astream.channels = 2
            astream.format = "s16"

            self.assertEqual(vstream.index, 0)
            self.assertEqual(astream.index, 1)

            vframe = VideoFrame(320, 240, "yuv420p")
            vpacket = vstream.encode(vframe)[0]

            self.assertIs(vpacket.stream, vstream)
            self.assertEqual(vpacket.stream_index, 0)

            for i in range(10):
                if astream.frame_size != 0:
                    frame_size = astream.frame_size
                else:
                    # decoder didn't indicate constant frame size
                    frame_size = 1000
                aframe = AudioFrame("s16", "stereo", samples=frame_size)
                aframe.sample_rate = 48000
                apackets = astream.encode(aframe)
                if apackets:
                    apacket = apackets[0]
                    break

            self.assertIs(apacket.stream, astream)
            self.assertEqual(apacket.stream_index, 1)

    def test_set_id_and_time_base(self):
        with av.open(self.sandboxed("output.mov"), "w") as output:
            stream = output.add_stream("mp2")
            self.assertIn(stream, output.streams.audio)

            # set id
            self.assertEqual(stream.id, 0)
            stream.id = 1
            self.assertEqual(stream.id, 1)

            # set time_base
            self.assertEqual(stream.time_base, None)
            stream.time_base = Fraction(1, 48000)
            self.assertEqual(stream.time_base, Fraction(1, 48000))



import av

from .common import TestCase, fate_suite


class TestAudioFifo(TestCase):
    def test_data(self):

        container = av.open(fate_suite("audio-reference/chorusnoise_2ch_44kHz_s16.wav"))
        stream = container.streams.audio[0]

        fifo = av.AudioFifo()

        input_ = []
        output = []

        for i, packet in enumerate(container.demux(stream)):
            for frame in packet.decode():
                input_.append(bytes(frame.planes[0]))
                fifo.write(frame)
                for frame in fifo.read_many(512, partial=i == 10):
                    output.append(bytes(frame.planes[0]))
            if i == 10:
                break

        input_ = b"".join(input_)
        output = b"".join(output)
        min_len = min(len(input_), len(output))

        self.assertTrue(min_len > 10 * 512 * 2 * 2)
        self.assertTrue(input_[:min_len] == output[:min_len])

    def test_pts_simple(self):

        fifo = av.AudioFifo()

        iframe = av.AudioFrame(samples=1024)
        iframe.pts = 0
        iframe.sample_rate = 48000
        iframe.time_base = "1/48000"

        fifo.write(iframe)

        oframe = fifo.read(512)
        self.assertTrue(oframe is not None)
        self.assertEqual(oframe.pts, 0)
        self.assertEqual(oframe.time_base, iframe.time_base)

        self.assertEqual(fifo.samples_written, 1024)
        self.assertEqual(fifo.samples_read, 512)
        self.assertEqual(fifo.pts_per_sample, 1.0)

        iframe.pts = 1024
        fifo.write(iframe)
        oframe = fifo.read(512)
        self.assertTrue(oframe is not None)
        self.assertEqual(oframe.pts, 512)
        self.assertEqual(oframe.time_base, iframe.time_base)

        iframe.pts = 9999  # Wrong!
        self.assertRaises(ValueError, fifo.write, iframe)

    def test_pts_complex(self):

        fifo = av.AudioFifo()

        iframe = av.AudioFrame(samples=1024)
        iframe.pts = 0
        iframe.sample_rate = 48000
        iframe.time_base = "1/96000"

        fifo.write(iframe)
        iframe.pts = 2048
        fifo.write(iframe)

        oframe = fifo.read_many(1024)[-1]

        self.assertEqual(oframe.pts, 2048)
        self.assertEqual(fifo.pts_per_sample, 2.0)

    def test_missing_sample_rate(self):

        fifo = av.AudioFifo()

        iframe = av.AudioFrame(samples=1024)
        iframe.pts = 0
        iframe.time_base = "1/48000"

        fifo.write(iframe)

        oframe = fifo.read(512)

        self.assertTrue(oframe is not None)
        self.assertIsNone(oframe.pts)
        self.assertEqual(oframe.sample_rate, 0)
        self.assertEqual(oframe.time_base, iframe.time_base)

    def test_missing_time_base(self):

        fifo = av.AudioFifo()

        iframe = av.AudioFrame(samples=1024)
        iframe.pts = 0
        iframe.sample_rate = 48000

        fifo.write(iframe)

        oframe = fifo.read(512)

        self.assertTrue(oframe is not None)
        self.assertIsNone(oframe.pts)
        self.assertIsNone(oframe.time_base)
        self.assertEqual(oframe.sample_rate, iframe.sample_rate)



#!/usr/bin/env python3

from fractions import Fraction
import colorsys

import numpy as np

import av


(width, height) = (640, 360)
total_frames = 20
fps = 30

container = av.open("generate_video_with_pts.mp4", mode="w")

stream = container.add_stream("mpeg4", rate=fps)  # alibi frame rate
stream.width = width
stream.height = height
stream.pix_fmt = "yuv420p"

# ffmpeg time is complicated
# more at https://github.com/PyAV-Org/PyAV/blob/main/docs/api/time.rst
# our situation is the "encoding" one

# this is independent of the "fps" you give above
# 1/1000 means milliseconds (and you can use that, no problem)
# 1/2 means half a second (would be okay for the delays we use below)
# 1/30 means ~33 milliseconds
# you should use the least fraction that makes sense for you
stream.codec_context.time_base = Fraction(1, fps)

# this says when to show the next frame
# (increment by how long the current frame will be shown)
my_pts = 0  # [seconds]
# below we'll calculate that into our chosen time base

# we'll keep this frame around to draw on this persistently
# you can also redraw into a new object every time but you needn't
the_canvas = np.zeros((height, width, 3), dtype=np.uint8)
the_canvas[:, :] = (32, 32, 32)  # some dark gray background because why not
block_w2 = int(0.5 * width / total_frames * 0.75)
block_h2 = int(0.5 * height / 4)

for frame_i in range(total_frames):

    # move around the color wheel (hue)
    nice_color = colorsys.hsv_to_rgb(frame_i / total_frames, 1.0, 1.0)
    nice_color = (np.array(nice_color) * 255).astype(np.uint8)

    # draw blocks of a progress bar
    cx = int(width / total_frames * (frame_i + 0.5))
    cy = int(height / 2)
    the_canvas[
        cy - block_h2 : cy + block_h2, cx - block_w2 : cx + block_w2
    ] = nice_color

    frame = av.VideoFrame.from_ndarray(the_canvas, format="rgb24")

    # seconds -> counts of time_base
    frame.pts = int(round(my_pts / stream.codec_context.time_base))

    # increment by display time to pre-determine next frame's PTS
    my_pts += 1.0 if ((frame_i // 3) % 2 == 0) else 0.5
    # yes, the last frame has no "duration" because nothing follows it
    # frames don't have duration, only a PTS

    for packet in stream.encode(frame):
        container.mux(packet)

# finish it with a blank frame, so the "last" frame actually gets shown for some time
# this black frame will probably be shown for 1/fps time
# at least, that is the analysis of ffprobe
the_canvas[:] = 0
frame = av.VideoFrame.from_ndarray(the_canvas, format="rgb24")
frame.pts = int(round(my_pts / stream.codec_context.time_base))
for packet in stream.encode(frame):
    container.mux(packet)

# the time should now be 15.5 + 1/30 = 15.533

# without that last black frame, the real last frame gets shown for 1/30
# so that video would have been 14.5 + 1/30 = 14.533 seconds long

# Flush stream
for packet in stream.encode():
    container.mux(packet)

# Close the file
container.close()
